import sys
import time
import itertools

from QuICT.simulation import Simulator
from QuICT.simulation.state_vector import ConstantStateVectorSimulator
from QuICT.tools import Logger, LogFormat
from QuICT.tools.circuit_library import CircuitLib


logger = Logger("CLI_Benchmark", LogFormat.zero)


def benchmark(gpu_enable: bool):
    # Convert qubits from str into List[int]
    alg_list = ["adder", "grover", "maxcut"]
    max_qubits = 10
    circuit_library = CircuitLib()
    simulation = ConstantStateVectorSimulator()

    for alg in alg_list:
        circuits = circuit_library.get_algorithm_circuit(alg, qubits_interval=max_qubits)
        for cir in circuits:
            start_time = time.time()
            _ = simulation.run(cir)
            end_time = time.time()

            logger.info(f"Spending Time: {end_time - start_time}; with {cir.name}")

    max_qubits = list(range(10, 20, 4))
    max_qubits_dm = list(range(5, 10, 2))
    circuit_library = CircuitLib()

    circuits = circuit_library.get_random_circuit("ustc", qubits_interval=max_qubits)
    circuits_dm = circuit_library.get_random_circuit("ustc", qubits_interval=max_qubits_dm)

    backend_list = ["state_vector", "density_matrix", "unitary"]
    device_list = ["CPU"]
    if gpu_enable:
        device_list.append("GPU")

    for backend, device in itertools.product(backend_list, device_list):
        simulator = Simulator(device, backend)
        if backend == "state_vector" or circuit_path is not None:
            circuits_list = circuits
        else:
            circuits_list = circuits_dm

        for cir in circuits_list:
            start_time = time.time()
            _ = simulator.run(cir)
            end_time = time.time()

            logger.info(f"Spending Time: {end_time - start_time}; with {cir.name} and backend {backend}, {device}")


if __name__ == "__main__":
    enable_gpu = sys.argv[1]

    # if type_ == "algorithm":
    #     alg_benchmark()
    # else:
    #     simulation_benchmark(*sys.argv[2:])



from typing import Iterable, List

from QuICT.core import Circuit
from QuICT.tools.circuit_library.circuitlib import CircuitLib
from QuICT.qcda.optimization.template_optimization.template_matching.template_matching import (
    MatchingDAGCircuit, TemplateMatching)
from QuICT.qcda.optimization.template_optimization.template_matching.template_substitution import \
    TemplateSubstitution
from QuICT.qcda.utility import OutputAligner

from .template_matching.template_substitution import CircuitCostMeasure


class TemplateOptimization(object):
    """
    Template optimization algorithm.

    [1] Iten, R., Moyard, R., Metger, T., Sutter, D. and Woerner, S., 2020.
    Exact and practical pattern matching for quantum circuit optimization.
    `arXiv:1909.05270 <https://arxiv.org/abs/1909.05270>`
    """

    def __init__(
            self,
            template_max_width=None,
            template_max_size=2,
            template_max_depth=None,
            template_typelist=None,
            template_list=None,
            qubit_fixing_num=1,
            prune_step=3,
            prune_survivor_num=1,
    ):
        """
        Execute template optimization algorithm.

        Specify `template_max_width/template_max_size/template_max_depth/template_typelist`
        if you want to use templates in CircuitLib limiting size/width/depth/gate types.
        No limit if set to None. By default all templates of size 2 in CircuitLib are used.

        Specify `template_list` if you want to use customized templates.
        Setting `template_list` will invalidate
        `template_max_width/template_max_size/template_max_depth/template_typelist`.

        `qubit_fixing_num`, `prune_step`, `prune_survivor_num` are 3 heuristic parameters
        used in template matching. Their default values are recommended values.
            1. `qubit_fixing_num` the number of additional qubits explored when enumerating
                the qubit mapping (default value is 1).
            2. `prune_step`, `prune_survivor_num` are parameters for backward matching.
                Backward match will prune the search tree when depth=k * `prune_step`
                (k = 1, 2, ...) and at most `prune_survivor_num` maximal matching scenarios
                will survive (default values are 3, 1).

        Args:
            template_max_width(int): Limit on number of qubits of templates used.
            template_max_size(int): Limit on number of gates of templates used.
            template_max_depth(int): Limit on depth of templates used.
            template_typelist(Iterable[GateType]): Limit on gate types of templates used.
            template_list(List[Circuit]): List of templates used.
            qubit_fixing_num(int): heuristic parameter for qubit exploring
            prune_step(int): heuristic parameter for backward match
            prune_survivor_num(int): heuristic parameter for backward match
        """

        if template_list is None:
            template_list = CircuitLib().get_template_circuit(
                template_max_width,
                template_max_size,
                template_max_depth,
                template_typelist
            )

        self.template_list = template_list
        self.heuristics_qubits_param = [qubit_fixing_num]
        self.heuristics_backward_param = [prune_step, prune_survivor_num]
        self.cost_measure = CircuitCostMeasure(target_device='nisq')

    def __repr__(self):
        return f'TemplateOptimization(' \
               f'heuristics_qubits_param={self.heuristics_qubits_param}), ' \
               f'heuristics_backward_param={self.heuristics_backward_param})'

    @OutputAligner()
    def execute(self, circuit):
        """
        Execute template optimization algorithm.

        Args:
            circuit(Circuit): the circuit to be optimized

        Returns:
            Circuit: the optimized circuit
        """

        circ_dag = MatchingDAGCircuit(circuit)

        for template in self.template_list:
            template_dag = MatchingDAGCircuit(template)

            if template_dag.width > circ_dag.width:
                continue

            matches = TemplateMatching.execute(
                circ_dag,
                template_dag,
                self.heuristics_qubits_param,
                self.heuristics_backward_param
            )

            circ_dag = TemplateSubstitution.execute(circ_dag, template_dag, matches, self.cost_measure)

        return circ_dag.get_circuit()



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2020/6/3 10:14 上午
# @Author  : Peng Sirui
# @File    : partial_grover.py

import numpy as np

from QuICT.core import Circuit
from QuICT.core.gate import *
from QuICT.core.gate.backend import MCTOneAux
from QuICT.tools import Logger
from QuICT.tools.exception.core import *

logger = Logger("Grover-partial")


def calculate_r1_r2_one_target(N, K):
    # see https://arxiv.org/abs/quant-ph/0504157
    r1 = np.sqrt(N) * np.pi * 0.25 - np.sqrt(N / K) * np.sqrt(3 / 4)
    r2 = np.sqrt(N / K) * np.pi * (1 / 6)
    r1 = round(r1)
    r2 = round(r2)
    return r1, r2


class PartialGrover:
    """ partial grover search with one target

    https://arxiv.org/abs/quant-ph/0407122
    """

    def __init__(self, simulator) -> None:
        self.simulator = simulator

    def circuit(self, n, n_block, n_ancilla, oracle, measure=True):
        """ partial grover search with one target

        Args:
            n(int):         bits length of global address
            n_block(int):   bits length of block address
            n_ancilla(int): length of oracle working space. assume clean
            oracle(CompositeGate): the oracle that flip phase of target state.
                [0:n] is index qreg,
                [n:n+k] is ancilla
                also assume that it works in style of QCQI p249 6.2
            measure(bool): measure included or not

        Returns:
            int: the target address, big endian
        """
        K = 1 << n_block
        N = 1 << n
        r1, r2 = calculate_r1_r2_one_target(N, K)

        circuit = Circuit(n + n_ancilla + 1)
        index_q = list(range(n))
        oracle_q = list(range(n, n + n_ancilla))
        ancillia_q = [n + n_ancilla]
        # step 1
        for idx in index_q:
            H | circuit(idx)
        for i in range(r1):
            # global inversion about target
            oracle | circuit(index_q + oracle_q)
            # global inversion about average
            for idx in index_q:
                H | circuit(idx)
            for idx in index_q:
                X | circuit(idx)
            H | circuit(index_q[n - 1])
            MCTOneAux().execute(n + 1) | circuit(index_q + oracle_q[:1])
            H | circuit(index_q[n - 1])
            for idx in index_q:
                X | circuit(idx)
            for idx in index_q:
                H | circuit(idx)
        # step 2
        for i in range(r2):
            # global inversion about target
            oracle | circuit(index_q + oracle_q)
            # local inversion about average
            local_n = n - n_block
            local_index_q = [j for j in range(n_block, n_block + local_n)]
            for idx in local_index_q:
                H | circuit(idx)
            for idx in local_index_q:
                X | circuit(idx)
            H | circuit(local_index_q[local_n - 1])
            MCTOneAux().execute(local_n + 1) | circuit(local_index_q + oracle_q[:1])
            H | circuit(local_index_q[local_n - 1])
            for idx in local_index_q:
                X | circuit(idx)
            for idx in local_index_q:
                H | circuit(idx)
        # step 3
        H | circuit(ancillia_q[0])
        X | circuit(ancillia_q[0])
        oracle | circuit(index_q + [ancillia_q[0]] + oracle_q[1:])
        X | circuit(ancillia_q[0])
        H | circuit(ancillia_q[0])
        # controlled inversion about average
        for idx in index_q:
            CH | circuit([ancillia_q[0], idx])
        for idx in index_q:
            CX | circuit([ancillia_q[0], idx])
        MCTOneAux().execute(n + 2) | circuit([ancillia_q[0]] + index_q + oracle_q[:1])
        for idx in index_q:
            CX | circuit([ancillia_q[0], idx])
        for idx in index_q:
            CH | circuit([ancillia_q[0], idx])
        # Measure
        for idx in index_q:
            if measure:
                Measure | circuit(idx)
        logger.info(
            f"circuit width           = {circuit.width():4}\n" +
            f"global Grover iteration = {r1:4}\n" +
            f"local  Grover iteration = {r2:4}\n" +
            f"oracle  calls           = {r1+r2+1:4}\n" +
            f"other circuit size      = {circuit.size() - oracle.size()*(r1+r2):4}\n"
        )
        return circuit

    def run(self, n, n_block, n_ancilla, oracle, measure=True):
        simulator = self.simulator
        index_q = list(range(n))
        circuit = self.circuit(n, n_block, n_ancilla, oracle, measure)
        simulator.run(circuit)
        return int(circuit[index_q])



import os
import shutil

from QuICT.tools.cli.client import QuICTLocalManager


# Local Controller
local_job_manager = QuICTLocalManager()


def get_template(output_path: str):
    if not os.path.exists(output_path):
        os.makedirs(output_path)

    tempfile_path = os.path.join(
        os.path.dirname(__file__),
        "../template/quict_job.yml"
    )
    shutil.copy(tempfile_path, output_path)


def start_job(file: str):
    local_job_manager.start_job(file)


def stop_job(name: str):
    local_job_manager.stop_job(name)


def restart_job(name: str):
    local_job_manager.restart_job(name)


def delete_job(name: str):
    local_job_manager.delete_job(name)


def status_job(name: str):
    local_job_manager.status_job(name)


def list_jobs():
    local_job_manager.list_job()



import copy
import random
from typing import Dict, List, Tuple, Union

import networkx as nx
import numpy as np
from QuICT.core import *
from QuICT.core.gate import BasicGate, CompositeGate


class CircuitInfo:
    """DAG Representation of a quantum circuit."""

    def __init__(self, circ: Union[Circuit, CompositeGate], max_gate_num: int) -> None:
        self._qubit_num = circ.width()
        self._max_gate_num = max_gate_num
        q = circ.width()
        if isinstance(circ, (Circuit, CompositeGate)):
            self._gates: List[BasicGate] = copy.deepcopy(circ.gates)
        else:
            raise TypeError(
                "circ argument only supports Circuit/CompositeGate/List[BasicGate]"
            )

        self._graph = nx.DiGraph()
        for gid in range(len(self._gates)):
            self._graph.add_node(gid)

        self._first_layer_gates = None

        occupied = [-1 for _ in range(q)]
        self._bit2gid: List[List[int]] = [[] for _ in range(q)]
        """Qubit to all gates on it."""
        for gid, gate in enumerate(self._gates):
            bit_num = gate.controls + gate.targets
            assert bit_num <= 2, "Only 1 or 2 bit gates are supported."
            self._reset_cache()
            if bit_num == 1:
                a = gate.targ
                self._bit2gid[a].append(gid)
                if occupied[a] != -1:
                    self._graph.add_edge(occupied[a], gid)
                nx.set_node_attributes(self._graph, {gid: {"gid": gid}})
                occupied[a] = gid
            elif bit_num == 2:
                a, b = gate.cargs + gate.targs
                assert a != b
                # Position to Gate ID
                self._bit2gid[a].append(gid)
                self._bit2gid[b].append(gid)
                # DAG edges
                if occupied[a] != -1:
                    self._graph.add_edge(occupied[a], gid)
                if occupied[b] != -1:
                    self._graph.add_edge(occupied[b], gid)
                nx.set_node_attributes(self._graph, {gid: {"gid": gid}})
                occupied[a] = gid
                occupied[b] = gid

        self._removed_cnt = 0

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, copy.deepcopy(v, memo))
        return result

    def copy(self):
        return copy.deepcopy(self)

    def count_gate(self) -> int:
        return nx.number_of_nodes(self._graph)

    def _reset_cache(self):
        self._first_layer_gates = None

    @property
    def first_layer_gates(self) -> Dict[int, BasicGate]:
        if self._first_layer_gates is not None:
            return self._first_layer_gates
        answer = self._get_first_layer_gates()
        self._first_layer_gates = answer
        return answer

    def _get_first_layer_gates(self) -> Dict[int, BasicGate]:
        answer = {}
        for bit_stick in self._bit2gid:
            if not bit_stick:
                # Skip if empty
                continue
            gid = bit_stick[0]
            gate = self._gates[gid]
            bit_num = gate.controls + gate.targets
            if bit_num == 2:
                a, b = gate.cargs + gate.targs
                if self._bit2gid[a][0] == self._bit2gid[b][0] and (gid not in answer):
                    answer[gid] = self._gates[gid]
            elif bit_num == 1:
                a = gate.targ
                answer[gid] = self._gates[gid]
        return answer

    def eager_exec(
        self,
        logic2phy: List[int],
        topo_graph: nx.Graph,
        physical_circ: CompositeGate = None,
    ) -> int:
        """Eagerly remove all executable gates for now.

        Args:
            logic2phy (List[int]): Current logical to physical qubit mapping.
            topo_graph (nx.Graph): Physical topology graph.
            physical_circ (CompositeGate, optional): If set, executed gates are appended to it.

        Returns:
            int: Removed (2-bit) gate number.
        """
        remove_cnt = 0
        remove_any = True
        self._reset_cache()
        while remove_any:
            remove_any = False
            for gid, gate in self._get_first_layer_gates().items():
                if gate.controls + gate.targets == 1:
                    a = gate.targ
                    _a = logic2phy[a]
                    self._bit2gid[a].pop(0)
                    remove_any = True
                    self._graph.remove_node(gid)
                    if physical_circ is not None:
                        with physical_circ:
                            gate & [_a]
                elif gate.controls + gate.targets == 2:
                    a, b = gate.cargs + gate.targs
                    assert self._bit2gid[a][0] == self._bit2gid[b][0]
                    _a, _b = logic2phy[a], logic2phy[b]
                    if topo_graph.has_edge(_a, _b):
                        self._bit2gid[a].pop(0)
                        self._bit2gid[b].pop(0)
                        remove_cnt += 1
                        remove_any = True
                        self._graph.remove_node(gid)
                        if physical_circ is not None:
                            with physical_circ:
                                gate & [_a, _b]
        self._removed_cnt += remove_cnt
        return remove_cnt

    def remained_circ(self, logic2phy: List[int]) -> CompositeGate:
        gates = {}
        for bit_stick in self._bit2gid:
            for gid in bit_stick:
                if gid not in gates:
                    gates[gid] = self._gates[gid]
        cg = CompositeGate()
        for gate in gates.values():
            args = gate.cargs + gate.targs
            args = [logic2phy[arg] for arg in args]
            with cg:
                gate & args
        return cg

    def biased_random_swap(
        self,
        topo_dist: np.ndarray,
        logic2phy: List[int],
        exclude=tuple(),
        zero_shift: float = 0.1,
    ) -> Tuple[int, int]:
        candidates = []
        weights = []

        for i in range(self._qubit_num):
            for j in range(i + 1, self._qubit_num):
                if abs(topo_dist[i][j] - 1) > 1e-6:
                    continue
                if (i, j) in exclude or (j, i) in exclude:
                    continue
                candidates.append((i, j))
                next_logic2phy = copy.copy(logic2phy)
                next_logic2phy[i], next_logic2phy[j] = (
                    next_logic2phy[j],
                    next_logic2phy[i],
                )
                bias = 0
                for gate in self.first_layer_gates.values():
                    if gate.controls + gate.targets != 2:
                        continue
                    a, b = gate.cargs + gate.targs
                    _a, _b = logic2phy[a], logic2phy[b]
                    prev_d = topo_dist[_a][_b]
                    _a, _b = next_logic2phy[a], next_logic2phy[b]
                    next_d = topo_dist[_a][_b]
                    bias += prev_d - next_d
                if abs(bias) < 1e-6:
                    bias = zero_shift
                bias = max(0, bias)
                weights.append(bias)
        assert len(candidates) > 0
        if random.random() < 0.5:
            action = random.choices(population=candidates, weights=weights, k=1)[0]
        else:
            action = random.choices(population=candidates, k=1)[0]
        return action



from typing import List

from ._operator import Operator
from QuICT.core.utils import unique_id_generator
from QuICT.tools.exception.core import TypeError, CheckPointNoChildError


class CheckPoint(Operator):
    """ The CheckPoint is a sign of the circuit, the Composite Gate with the related
    CheckPointChild will be add into the sign indexes of the circuit.

    Example:
        Use the syntax "CheckPoint | Circuit" to add a CheckPointChild into the Circuit.
        A circuit may have more than one CheckPoints.
    """
    @property
    def uid(self):
        """ Unique Identity Number. Used for mapping CheckPoint and CheckPointChild. """
        return self._uid

    @property
    def position(self):
        """ The related index in the circuit. """
        return self._pos

    @position.setter
    def position(self, shift: int):
        """ Set the related index of the circuit. """
        assert isinstance(shift, int), TypeError("CheckPoint.position.shift", "int", type(shift))
        self._pos += shift

    def __init__(self):
        super().__init__(targets=1)
        self._uid = unique_id_generator()
        self._pos = -1

    def get_child(self, shift: int = 0):
        """ Generate its CheckPointChild, which has the same uid. """
        assert isinstance(shift, int), TypeError("CheckPoint.get_child.shift", "int", type(shift))
        return CheckPointChild(self.uid, shift)

    def match(self, uid: str) -> bool:
        """ Comparsion the uid with self. """
        return uid == self._uid

    def __or__(self, targets):
        """ Add CheckPoint into circuit, and set position to be the end of current circuit. """
        super().__or__(targets)
        self._pos = targets.size()


class CheckPointChild(Operator):
    """ The child for CheckPoint, used with the CompositGate. The purpose of the CheckPointChild is
    to add CompositeGate into the target index of the circuit.

    Example:
        Use the syntax "CheckPointChild | CompositeGate" to add a CheckPointChild into the CompositeGate.
    """
    @property
    def uid(self):
        """ Unique identity number. Used for mapping CheckPoint and CheckPointChild. """
        return self._uid

    @property
    def shift(self) -> int:
        """ The shift distance of the CheckPoint position. """
        return self._shift

    def __init__(self, uid: str, shift: int = 0):
        """ Initial a CheckPointChild.

        Args:
            uid (str): The unique identity number, same with its CheckPoint.
            shift (int, optional): The shift distance of the CheckPoint position. Defaults to 0.
        """
        super().__init__(targets=1)
        self._uid = uid
        self._shift = shift

    def find_position(self, checkpoints: List[CheckPoint]) -> int:
        """ return the position of the mappin CheckPoint. """
        pos = -1
        for cp in checkpoints:
            if cp.match(self._uid):
                pos = cp.position
                cp.position = self._shift

        if pos == -1:
            raise CheckPointNoChildError("Cannot find parent checkpoint for current checkpointchild.")

        return pos



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2021/6/30 下午5:47
# @Author  : Kaiqi Li
# @File    : algorithm loader

import importlib


DEVICE_LIST = ["CPU", "GPU"]
BASED_LINALG = [
    "dot", "tensor", "MatrixTensorI", "MatrixPermutation", "VectorPermutation",
    "matrix_dot_vector", "simple_vp", "qubit_vp"
]


class LinAlgLoader:
    """
    The Algorithm class with used to load all required algorithm, including based linear algorithm, gate-based matrix
    dot vector algorithm, and algorithms for multi-GPUs.

    Args:
        device(str): one of ["GPU", "CPU"].
        enable_gate_kernel(bool): loading gate-based matrix dot vector algorithm if True.
        enable_multigpu_gate_kernel(bool): loading the required algorithms of multi-GPUs if True.
    """
    def __init__(self, device: str, enable_gate_kernel: bool = False, enable_multigpu_gate_kernel: bool = False):
        if device not in DEVICE_LIST:
            raise KeyError(f"Not supported the given device, please choice one of {DEVICE_LIST}")

        if device == "CPU":
            linalg_lib = importlib.import_module('QuICT.ops.linalg.cpu_calculator')
        else:
            linalg_lib = importlib.import_module('QuICT.ops.linalg.gpu_calculator')

        for attr, value in linalg_lib.__dict__.items():
            if attr in BASED_LINALG:
                self.__dict__[attr] = value

        if enable_gate_kernel:
            gate_lib = importlib.import_module('QuICT.ops.gate_kernel.gate_function')
            GATE_KERNEL_FUNCTIONS = gate_lib.__dict__["__outward_functions"]

            for attr, value in gate_lib.__dict__.items():
                if attr in GATE_KERNEL_FUNCTIONS:
                    self.__dict__[attr] = value

        if enable_multigpu_gate_kernel:
            proxy_lib = importlib.import_module('QuICT.ops.gate_kernel.multigpu_gate_func')
            PROXY_GATE_FUNCTIONS = proxy_lib.__dict__["__outward_functions"]

            for attr, value in proxy_lib.__dict__.items():
                if attr in PROXY_GATE_FUNCTIONS:
                    self.__dict__[attr] = value



import json
import jwt
import os
import functools
from flask import request, Response

from .encrypt_manager import EncryptManager
from .sql_controller import SQLManger


__SALT = "TestForQuICT"


def format_job_dict(job_dict: dict):
    for key, value in job_dict.items():
        if not isinstance(value, (str, int, float, bytes)):
            job_dict[key] = json.dumps(value)

    return job_dict


def request_validation(login: bool = False, register: bool = False):
    """Check JWT validity and do data decryption before getting into the actual logistic.
    Args:
        func:
    Returns:
        None.
    """

    def decorator(func):
        @functools.wraps(func)
        def with_valid(*args, **kwargs):
            sql_conn = SQLManger()
            # Get jwt_token and its payload
            encrypt = EncryptManager()
            jwt_token = request.headers.get('Authorization')
            if jwt_token and jwt_token.startswith('Bearer '):
                payload = jwt.decode(jwt_token[7:], __SALT, ["HS256"])
            else:
                payload = None

            username = payload.get("username", None)
            if register and sql_conn.validate_user(username):
                return create_response(username, __SALT, {'error': "Existed users!"})

            if not register and not sql_conn.validate_user(username):
                return create_response(username, __SALT, {'error': "unauthorized user"})

            kwargs['username'] = username
            aes_key = payload.get('aes_key')
            encrypted_passwd = sql_conn.get_password(username)[:16] if not login and not register else \
                __SALT

            data = request.data
            if data != b'':
                decrypted_aeskey = encrypt.decryptedmsg(aes_key, encrypted_passwd, True)
                decrypted_data = encrypt.decryptedmsg(data, decrypted_aeskey)
                kwargs['json_dict'] = format_job_dict(json.loads(decrypted_data))

            try:
                return_data = func(*args, **kwargs)
            except Exception as e:
                return_data = {'error': repr(e)}
            finally:
                # create header for response
                return create_response(username, encrypted_passwd, return_data)

        return with_valid

    return decorator


def create_response(username: str, password: str, json_dict: dict):
    encrypt = EncryptManager()
    aes_key = os.urandom(16)

    payload = {
        'username': username,
        'aes_key': encrypt.encryptedmsg(aes_key, password).decode('ascii')
    }

    jwt_token = jwt.encode(
        payload=payload,
        key=__SALT,
        algorithm="HS256"
    )

    if json_dict is not None:
        json_dict = encrypt.encryptedmsg(json.dumps(json_dict), aes_key)

    response = Response(response=json_dict)
    response.headers = {"Authorization": f"Bearer {jwt_token}"}

    return response



from flask import Flask

from job import job_blueprint
from env import env_blueprint


app = Flask(__name__)
app.register_blueprint(blueprint=job_blueprint)
app.register_blueprint(blueprint=env_blueprint)


if __name__ == "__main__":
    app.run(port=5000, debug=True)



from typing import *

import numpy as np

from QuICT.core import *
from QuICT.core.gate import build_gate, BasicGate, CompositeGate, GateType
from QuICT.qcda.utility import OutputAligner
from .utility import *
from .graph import *
from .edge_coloring import EdgeColoring
from .block_ldu_decompose import BlockLDUDecompose


class CnotWithoutAncilla(object):
    @OutputAligner()
    def execute(
            self,
            circuit_segment: Union[Circuit, CompositeGate]
    ) -> CompositeGate:
        if isinstance(circuit_segment, Circuit):
            gates = circuit_segment.gates
            n = len(circuit_segment.qubits)
        elif isinstance(circuit_segment, CompositeGate):
            gates = circuit_segment
            n = circuit_segment.circuit_width()
        else:
            raise Exception("Only accept Circuit/CompositeGate!")

        mat = np.eye(n, dtype=bool)
        for gate in gates:
            gate: BasicGate
            c = gate.carg
            t = gate.targ
            mat[t, :] ^= mat[c, :]

        composite_gate = CompositeGate()

        parallel_elimination = self.matrix_run(mat)
        parallel_elimination.reverse()
        for level in parallel_elimination:
            for c, t in level:
                _cx = build_gate(GateType.cx, [c, t])
                composite_gate.append(_cx)
        return composite_gate

    @classmethod
    def matrix_run(cls, mat: np.ndarray) \
            -> List[List[Tuple[int, int]]]:
        """
        Get parallel row elimination of a boolean invertible matrix.

        Args:
            mat (np.ndarray): A boolean invertible matrix

        Returns:
            List[List[Tuple[int,int]]]: Parallel row elimination. In each depth level there are
            non-overlapping row eliminations.
        """
        if len(mat.shape) != 2 or mat.shape[0] != mat.shape[1]:
            raise Exception("Must use a square matrix!")
        if mat.dtype != bool:
            raise Exception("Must use a boolean matrix!")

        return cls.__matrix_run(mat)

    @classmethod
    def small_matrix_run(cls, mat: np.ndarray) \
            -> List[List[Tuple[int, int]]]:
        n = mat.shape[0]
        if n == 1:
            return [[]]
        elif n == 2:
            # Enumerate
            if np.allclose(
                    mat,
                    np.array([
                        [1, 0],
                        [0, 1]
                    ], dtype=bool)
            ):  # :(
                return [[]]
            elif np.allclose(
                    mat,
                    np.array([
                        [1, 0],
                        [1, 1]
                    ], dtype=bool)
            ):  # :P
                return [[(0, 1)]]
            elif np.allclose(
                    mat,
                    np.array([
                        [0, 1],
                        [1, 0]
                    ], dtype=bool)
            ):  # :b
                return [[(0, 1)], [(1, 0)], [(0, 1)]]
            elif np.allclose(
                    mat,
                    np.array([
                        [0, 1],
                        [1, 1]
                    ], dtype=bool)
            ):  # :D
                return [[(1, 0)], [(0, 1)]]
            elif np.allclose(
                    mat,
                    np.array([
                        [1, 1],
                        [0, 1]
                    ], dtype=bool)
            ):  # :X
                return [[(1, 0)]]
            elif np.allclose(
                    mat,
                    np.array([
                        [1, 1],
                        [1, 0]
                    ], dtype=bool)
            ):  # :O
                return [[(0, 1)], [(1, 0)]]
        else:
            raise Exception("Must use matrix size <= 2.")

    @classmethod
    def __matrix_run(cls, mat: np.ndarray) \
            -> List[List[Tuple[int, int]]]:
        """
        Get parallel row elimination of a boolean invertible matrix.
        No shape & data type checks for inner methods.

        Args:
            mat (np.ndarray): A boolean invertible matrix

        Returns:
            List[List[Tuple[int,int]]]: Parallel row elimination. In each depth level there are
            non-overlapping row eliminations.
        """

        n = mat.shape[0]

        if n <= 2:
            return cls.small_matrix_run(mat)

        l_, d_, u_, remapping = BlockLDUDecompose.run(mat)

        # Implement remapping using swap operations(3-depth CNOT).

        """
        Notes:
        P @ M == L @ D @ U, where P == P^{-1}.
        So, M == P @ L @ D @ U. We sequentially eliminate P, L, D, U.
        If we reverse the eliminations' order to get a row transform sequence,
        changing an I into M, then that is exactly what we need for CNOT circuit.

        """

        parallel_elimination: List[List[Tuple[int, int]]] = []

        r_p_e = cls.remapping_run(remapping)
        parallel_elimination.extend(r_p_e)

        l_p_e = cls.triangular_matrix_run(l_, is_lower_triangular=True)
        if l_p_e != [[]]:
            parallel_elimination.extend(l_p_e)

        d_p_e = cls.block_diagonal_matrix_run(d_)
        if d_p_e != [[]]:
            parallel_elimination.extend(d_p_e)

        u_p_e = cls.triangular_matrix_run(u_, is_lower_triangular=False)
        if u_p_e != [[]]:
            parallel_elimination.extend(u_p_e)

        return parallel_elimination

    @classmethod
    def triangular_matrix_run(cls, mat: np.ndarray, is_lower_triangular: bool) \
            -> List[List[Tuple[int, int]]]:
        """
        Get parallel row elimination of a triangular matrix by bipartite edge coloring.


        Args:
            mat (np.ndarray): A triangular boolean matrix with block identity diagonal entries.
            is_lower_triangular (bool): Is the input triangular matrix has all its non-zero
                entries in lower part. If this parameter is false, all non-zero entries should
                be located in upper part.

        Returns:
            List[List[Tuple[int,int]]]: Parallel row elimination. In each depth level there are
            non-overlapping row eliminations.
        """
        n = mat.shape[0]
        s_size = n // 2
        t_size = n - s_size

        # Always use the identity sub matrix as the `left` in bipartite
        if is_lower_triangular:
            left = [i for i in range(s_size)]
            right = [i + s_size for i in range(t_size)]
        else:
            left = [i for i in range(t_size)]
            right = [i + t_size for i in range(s_size)]

        # Construct bipartite
        bipartite = Bipartite(left, right)
        # Add edges
        if is_lower_triangular:
            for j in range(s_size):
                for i in range(s_size, n):
                    if mat[i, j]:
                        bipartite.add_edge(j, i)
                        bipartite.add_edge(i, j)
        else:
            for j in range(s_size, n):
                for i in range(s_size):
                    if mat[i, j]:
                        bipartite.add_edge(j - s_size, i + t_size)
                        bipartite.add_edge(i + t_size, j - s_size)

        colored_bipartite = EdgeColoring.get_edge_coloring(bipartite)
        max_deg = colored_bipartite.get_max_degree()
        parallel_elimination: List[List[Tuple[int, int]]] = [[] for _ in range(max_deg)]

        # Iterate over left vertices to check edges with the same color
        for node in colored_bipartite.left:
            eid = colored_bipartite.head[node]
            while eid != -1:
                edge = colored_bipartite.edges[eid]
                color_index = edge.color
                if is_lower_triangular:
                    c = edge.start
                    t = edge.end
                else:
                    c = edge.start + s_size
                    t = edge.end - t_size
                row_elimination = (c, t)
                parallel_elimination[color_index].append(row_elimination)
                eid = edge.next

        return parallel_elimination

    @classmethod
    def block_diagonal_matrix_run(cls, mat: np.ndarray) \
            -> List[List[Tuple[int, int]]]:
        """
        Get parallel row elimination of a block diagonal matrix by bipartite edge coloring.


        Args:
            mat (np.ndarray): A block diagonal boolean matrix.

        Returns:
            List[List[Tuple[int,int]]]: Parallel row elimination. In each depth level there are
            non-overlapping row eliminations.
        """
        n = mat.shape[0]
        s_size = n // 2

        u1 = mat[:s_size, :s_size]
        u2 = mat[s_size:, s_size:]

        u1_parallel_elimination = cls.__matrix_run(u1)
        u2_parallel_elimination = cls.__matrix_run(u2)

        u1_p_l = len(u1_parallel_elimination)
        u2_p_l = len(u2_parallel_elimination)
        for lv in u2_parallel_elimination:
            for idx, r in enumerate(lv):
                lv[idx] = (r[0] + s_size, r[1] + s_size)
        p_l = max(u1_p_l, u2_p_l)

        parallel_elimination: List[List[Tuple[int, int]]] = [[] for _ in range(p_l)]

        for i in range(p_l):
            if i < u1_p_l:
                parallel_elimination[i].extend(u1_parallel_elimination[i])
            if i < u2_p_l:
                parallel_elimination[i].extend(u2_parallel_elimination[i])

        return parallel_elimination

    @classmethod
    def remapping_run(cls, remapping: List[int]) -> List[List[Tuple[int, int]]]:
        s_size = len(remapping) // 2

        parallel_elimination: List[List[Tuple[int, int]]] = [[], [], []]

        for i in range(s_size):
            if remapping[i] >= s_size:
                x = remapping[i]
                y = i
                parallel_elimination[0].append((x, y))
                parallel_elimination[1].append((y, x))
                parallel_elimination[2].append((x, y))

        return parallel_elimination



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2021/3/20 4:53 下午
# @Author  : Han Yu
# @File    : gate_transform.py

import numpy as np

from QuICT.core import Circuit
from QuICT.core.gate import *
from QuICT.qcda.synthesis.gate_transform.special_set import *
from QuICT.qcda.utility import OutputAligner


class GateTransform(object):
    def __init__(self, instruction_set=USTCSet):
        """
        Args:
            instruction_set(InstructionSet): the goal instruction set
        """
        self.instruction_set = instruction_set

    @OutputAligner()
    def execute(self, circuit):
        """ equivalently transfrom circuit into goal instruction set

        The algorithm will try two possible path, and return a better result:
        1. make continuous local two-qubit gates into SU(4), then decomposition with goal instruction set
        2. transform the two-qubit gate into instruction set one by one, then one-qubit gate

        Args:
            circuit(Circuit/CompositeGate): the circuit to be transformed

        Returns:
            CompositeGate: the equivalent compositeGate with goal instruction set
        """
        gates = CompositeGate()
        if isinstance(circuit, CompositeGate):
            gates.extend(circuit)
        elif isinstance(circuit, Circuit):
            gates.extend(circuit.gates)
        else:
            raise TypeError("Invalid input for GateTransform")

        gates.gate_decomposition()
        gates = self.two_qubit_transform(gates)
        gates = self.one_qubit_transform(gates)
        return gates

    def one_qubit_transform(self, gates):
        gates_tran = CompositeGate()
        unitaries = [np.identity(2, dtype=np.complex128) for _ in range(gates.width())]
        for gate in gates:
            if gate.targets + gate.controls == 2:
                targs = gate.cargs + gate.targs
                for targ in targs:
                    gates_transformed = self.instruction_set.one_qubit_rule(Unitary(unitaries[targ]) & targ)
                    if gates_transformed.width() == 0:
                        local_matrix = np.eye(2)
                    else:
                        local_matrix = gates_transformed.matrix(local=True)
                    phase = np.angle(np.dot(unitaries[targ], np.linalg.inv(local_matrix))[0][0])
                    if (
                        not np.isclose(np.mod(phase, 2 * np.pi), 0) and
                        not np.isclose(np.mod(phase, 2 * np.pi), 2 * np.pi)
                    ):
                        gates_transformed.append(GPhase(phase) & targ)
                    gates_tran.extend(gates_transformed)
                    unitaries[targ] = np.identity(2, dtype=np.complex128)
                gates_tran.append(gate)
            else:
                unitaries[gate.targ] = np.dot(gate.matrix, unitaries[gate.targ])
        for i in range(gates.width()):
            gates_transformed = self.instruction_set.one_qubit_rule(Unitary(unitaries[i]) & i)
            if gates_transformed.width() == 0:
                local_matrix = np.eye(2)
            else:
                local_matrix = gates_transformed.matrix(local=True)
            phase = np.angle(np.dot(unitaries[i], np.linalg.inv(local_matrix))[0][0])
            if (
                not np.isclose(np.mod(phase, 2 * np.pi), 0) and
                not np.isclose(np.mod(phase, 2 * np.pi), 2 * np.pi)
            ):
                gates_transformed.append(GPhase(phase) & i)
            gates_tran.extend(gates_transformed)
        return gates_tran

    def two_qubit_transform(self, gates):
        gates_tran = CompositeGate()
        for gate in gates:
            if gate.targets + gate.controls > 2:
                raise Exception("gate_transform only support 2-qubit and 1-qubit gate now.")
            if gate.type != self.instruction_set.two_qubit_gate and gate.targets + gate.controls == 2:
                rule = self.instruction_set.select_transform_rule(gate.type)
                gates_tran.extend(rule(gate))
            else:
                gates_tran.append(gate)
        return gates_tran



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2021/6/29 下午5:35
# @Author  : Kaiqi Li
# @File    : gpu_calculator

import math
from typing import List, Union
import numpy as np
import cupy as cp


def dot(A, B, gpu_out: bool = False, sync: bool = True):
    """ dot matrix A and matrix B

    Args:
        A(np.array<np.complex>): the matrix A
        B(np.array<np.complex>): the matrix B
        gpu_out(bool): return result from GPU into CPU

    Returns:
        np.array<np.complex>: A * B
    """
    assert (A.shape[1] == B.shape[0])

    # Data in GPU.
    gpu_A = cp.array(A) if type(A) is np.ndarray else A
    gpu_B = cp.array(B) if type(B) is np.ndarray else B

    gpu_result = cp.dot(gpu_A, gpu_B)

    if sync:
        cp.cuda.Device().synchronize()

    if gpu_out:
        return gpu_result.get()

    return gpu_result


tensor_single_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void tensorsingle(complex<float>* x, complex<float>* y, complex<float>* out, int cx, int ry, int cy) {
        int out_id = blockDim.x * blockIdx.x + threadIdx.x;
        int out_width = cx * cy;
        int y_size = ry * cy;
        int x_id = (out_id/out_width)/ry * cx + (out_id%out_width) / cy;
        int y_id = (out_id/out_width)%ry * cy + (out_id%out_width) % cy;
        out[out_id] = x[x_id] * y[y_id];
    }
    ''', 'tensorsingle')


tensor_double_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void tensordouble(complex<double>* x, complex<double>* y, complex<double>* out, int cx, int ry, int cy) {
        int out_id = blockDim.x * blockIdx.x + threadIdx.x;
        int out_width = cx * cy;
        int y_size = ry * cy;
        int x_id = (out_id/out_width)/ry * cx + (out_id%out_width) / cy;
        int y_id = (out_id/out_width)%ry * cy + (out_id%out_width) % cy;
        out[out_id] = x[x_id] * y[y_id];
    }
    ''', 'tensordouble')


def tensor(A, B, gpu_out: bool = False, sync: bool = True):
    """ tensor A and B

    Args:
        A(np.array<np.complex>): the matrix A
        B(np.array<np.complex>): the matrix B
        gpu_out(bool): return result from GPU into CPU

    Returns:
        np.array<np.complex>: the tensor result A ⊗ B
    """
    # Data in GPU.
    gpu_A = cp.array(A) if type(A) is np.ndarray else A
    gpu_B = cp.array(B) if type(B) is np.ndarray else B

    row_a, row_b = A.shape[0], B.shape[0]
    col_a = 1 if A.ndim == 1 else A.shape[1]
    col_b = 1 if B.ndim == 1 else B.shape[1]

    gpu_result = cp.empty((row_a * row_b, col_a * col_b), dtype=A.dtype)
    core_number = gpu_result.size
    kernel_function = tensor_single_kernel if A.dtype == np.complex64 else tensor_double_kernel
    kernel_function(
        (math.ceil(core_number / 1024),),
        (min(1024, core_number),),
        (gpu_A, gpu_B, gpu_result, cp.int32(col_a), cp.int32(row_b), cp.int32(col_b))
    )

    if sync:
        cp.cuda.Device().synchronize()

    if gpu_out:
        return gpu_result.get()

    return gpu_result


matrixt_single_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void matrix_tensorI_single(const complex<float>* A, complex<float>* out, int n, int m, int rx, int cx) {
        int tid = blockDim.x * blockIdx.x + threadIdx.x;
        int x_id = tid/cx;
        int y_id = (x_id/(rx*m))*(m*cx) + (tid%cx)*m + (x_id%(rx*m))%m;
        int out_xid = (x_id%(rx*m))/m;
        int out_yid = tid%cx;
        out[x_id*(cx*n*m) + y_id] = A[out_xid*cx + out_yid];
    }
    ''', 'matrix_tensorI_single')


matrixt_double_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void matrix_tensorI_double(const complex<double>* A, complex<double>* out, int n, int m, int rx, int cx) {
        int tid = blockDim.x * blockIdx.x + threadIdx.x;
        int x_id = tid/cx;
        int y_id = (x_id/(rx*m))*(m*cx) + (tid%cx)*m + (x_id%(rx*m))%m;
        int out_xid = (x_id%(rx*m))/m;
        int out_yid = tid%cx;
        out[x_id*(cx*n*m) + y_id] = A[out_xid*cx + out_yid];
    }
    ''', 'matrix_tensorI_double')


def MatrixTensorI(A, n, m, gpu_out: bool = False, sync: bool = True):
    """ tensor I^n and A and I^m

    Args:
        A(np.array<np.complex>): the matrix A
        n(int): the index of indentity
        m(int): the index of indentity
        gpu_out(bool): return result from GPU into CPU

    Returns:
        np.array<np.complex>: the tensor result I^n ⊗ A ⊗ I^m
    """
    row_a, col_a = A.shape

    # Data in GPU.
    precision = A.dtype
    gpu_A = cp.array(A) if type(A) is np.ndarray else A

    gpu_result = cp.zeros((row_a * n * m, col_a * n * m), dtype=precision)
    core_number = gpu_A.size * n * m
    kernel_function = matrixt_single_kernel if A.dtype == np.complex64 else matrixt_double_kernel
    kernel_function(
        (math.ceil(core_number / 1024),),
        (min(1024, core_number),),
        (gpu_A, gpu_result, cp.int32(n), cp.int32(m), cp.int32(row_a), cp.int32(col_a))
    )

    if sync:
        cp.cuda.Device().synchronize()

    if gpu_out:
        return gpu_result.get()

    return gpu_result


vectorp_single_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void vector_single_permutation(const complex<float>* x, complex<float>* y, int* mapping, int m) {
        int tid = blockDim.x * blockIdx.x + threadIdx.x;
        int xid = 0;
        int l = 0;
        for(int i = 0; i < m; i++){
            if ((i != m - 1) && ((mapping[i] + 1) == mapping[i+1])){
                l += 1;
            }else{
                xid |= ((tid >> (m - 1 - mapping[i])) & ((1 << (l + 1)) - 1)) << (m - 1 - i);
                l = 0;
            }
        }
        y[tid] = x[xid];
    }
    ''', 'vector_single_permutation')


vectorp_double_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void vector_double_permutation(const complex<double>* x, complex<double>* y, int* mapping, int m) {
        int tid = blockDim.x * blockIdx.x + threadIdx.x;
        int xid = 0;
        int l = 0;
        for(int i = 0; i < m; i++){
            if ((i != m - 1) && ((mapping[i] + 1) == mapping[i+1])){
                l += 1;
            }else{
                xid |= ((tid >> (m - 1 - mapping[i])) & ((1 << (l + 1)) - 1)) << (m - 1 - i);
                l = 0;
            }
        }
        y[tid] = x[xid];
    }
    ''', 'vector_double_permutation')


def VectorPermutation(A, mapping, changeInput: bool = False, gpu_out: bool = False, sync: bool = True):
    """ permutaion A with mapping, inplace

    Args:
        A(np.array<np.complex>): the matrix A.
        mapping(np.array<int>): the qubit mapping.
        changeInput(bool): whether changes in A.
        gpu_out(bool): return result from GPU.

    Returns:
        np.array<np.complex>: the result of Permutation
    """
    row_a, n = A.shape[0], mapping.shape[0]
    if not row_a == 1 << n:
        raise IndexError("Indices do not match!")

    if mapping.dtype == np.int64:
        mapping = mapping.astype(np.int32)

    # data in GPU
    gpu_A = cp.array(A) if type(A) is np.ndarray else A
    gpu_mapping = cp.array(mapping)
    gpu_result = cp.empty_like(gpu_A)
    core_number = gpu_result.size
    kernel_function = vectorp_single_kernel if A.dtype == np.complex64 else vectorp_double_kernel
    kernel_function(
        (math.ceil(core_number / 1024),),
        (min(1024, core_number),),
        (gpu_A, gpu_result, gpu_mapping, cp.int32(n))
    )

    if changeInput:
        A[:] = gpu_result.get() if type(A) is np.ndarray else gpu_result

    if sync:
        cp.cuda.Device().synchronize()

    if gpu_out:
        return gpu_result.get()

    return gpu_result


matrixp_single_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void matrix_single_permutation(const complex<float>* x, complex<float>* y, int* mapping, int m) {
        int tid = blockDim.x * blockIdx.x + threadIdx.x;
        int len = 1 << m;
        int rx = tid/len;
        int cx = tid%len;
        int rtemp = 0;
        int ctemp = 0;
        int l = 1;
        for(int i = 0; i < m; i++){
            if ((i != m - 1) && ((mapping[i] | 1) == mapping[i+1])){
                l = (l << 1) | 1;
            }else{
                rtemp |= ((rx >> (m - 1 - mapping[i])) & l) << (m - 1 - i);
                ctemp |= ((cx >> (m - 1 - mapping[i])) & l) << (m - 1 - i);
                l = 1;
            }
        }
        y[tid] = x[rtemp*len + ctemp];
    }
    ''', 'matrix_single_permutation')


matrixp_double_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void matrix_double_permutation(const complex<double>* x, complex<double>* y, int* mapping, int m) {
        int tid = blockDim.x * blockIdx.x + threadIdx.x;
        int len = 1 << m;
        int rx = tid/len;
        int cx = tid%len;
        int rtemp = 0;
        int ctemp = 0;
        int l = 1;
        for(int i = 0; i < m; i++){
            if ((i != m - 1) && ((mapping[i] | 1) == mapping[i+1])){
                l = (l << 1) | 1;
            }else{
                rtemp |= ((rx >> (m - 1 - mapping[i])) & l) << (m - 1 - i);
                ctemp |= ((cx >> (m - 1 - mapping[i])) & l) << (m - 1 - i);
                l = 1;
            }
        }
        y[tid] = x[rtemp*len + ctemp];
    }
    ''', 'matrix_double_permutation')


def MatrixPermutation(A, mapping, changeInput: bool = False, gpu_out: bool = False, sync: bool = True):
    """ permute mat with mapping, inplace

    Args:
        A(np.array<np.complex>): the matrix A.
        mapping(np.array<int>): the qubit mapping.
        changeInput(bool): whether changes in A.
        gpu_out(bool): return result from GPU.
    """
    row_a, n = A.shape[0], mapping.shape[0]
    if not row_a == 1 << n:
        raise IndexError("Indices do not match!")

    if mapping.dtype == np.int64:
        mapping = mapping.astype(np.int32)

    # data in GPU
    gpu_A = cp.array(A) if type(A) is np.ndarray else A
    gpu_mapping = cp.array(mapping)
    gpu_result = cp.empty_like(gpu_A)
    core_number = gpu_result.size
    kernel_function = matrixp_single_kernel if A.dtype == np.complex64 else matrixp_double_kernel
    kernel_function(
        (math.ceil(core_number / 1024),),
        (min(1024, core_number),),
        (gpu_A, gpu_result, gpu_mapping, cp.int32(n))
    )

    if changeInput:
        A[:, :] = gpu_result.get() if type(A) is np.ndarray else gpu_result

    if sync:
        cp.cuda.Device().synchronize()

    if gpu_out:
        return gpu_result.get()

    return gpu_result


matrix_dot_vector_single_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void matrix_dot_vector_single(
        const complex<float>* mat,
        int mat_bit,
        int mat_len,
        complex<float>* vec,
        int* affect_args,
        int* aff_argsorts,
        complex<float>* anc
    ){
        int tid = blockDim.x * blockIdx.x + threadIdx.x;
        int other = tid & ((1 << aff_argsorts[0]) - 1);
        int gw = tid >> aff_argsorts[0] << (aff_argsorts[0] + 1);
        for(int i = 1; i < mat_bit; i++){
            other += gw & ((1 << aff_argsorts[i]) - (1 << aff_argsorts[i - 1]));
            gw = gw >> aff_argsorts[i] << (aff_argsorts[i] + 1);
        }
        other += gw;

        for(int i = 0; i < mat_len; i++){
            int now = other;
            for(int k = 0; k < mat_bit; k++){
                if (i & (1 << k)){
                    now += 1 << affect_args[mat_bit - 1 - k];
                }
            }
            anc[now] = 0;
            for(int k = 0; k < mat_len; k++){
                int shift = other;
                for(int l = 0; l < mat_bit; l++){
                    if (k & (1 << l)){
                        shift += 1 << affect_args[mat_bit - 1 - l];
                    }
                }
                anc[now] += mat[i*mat_len + k] * vec[shift];
            }
        }

    }
    ''', 'matrix_dot_vector_single')


matrix_dot_vector_double_kernel = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void matrix_dot_vector_double(
        const complex<double>* mat,
        int mat_bit,
        int mat_len,
        complex<double>* vec,
        int* affect_args,
        int* aff_argsorts,
        complex<double>* anc
    ){
        int tid = blockDim.x * blockIdx.x + threadIdx.x;
        int other = tid & ((1 << aff_argsorts[0]) - 1);
        int gw = tid >> aff_argsorts[0] << (aff_argsorts[0] + 1);
        for(int i = 1; i < mat_bit; i++){
            other += gw & ((1 << aff_argsorts[i]) - (1 << aff_argsorts[i - 1]));
            gw = gw >> aff_argsorts[i] << (aff_argsorts[i] + 1);
        }
        other += gw;

        for(int i = 0; i < mat_len; i++){
            int now = other;
            for(int k = 0; k < mat_bit; k++){
                if (i & (1 << k)){
                    now += 1 << affect_args[mat_bit - 1 - k];
                }
            }
            for(int k = 0; k < mat_len; k++){
                int shift = other;
                for(int l = 0; l < mat_bit; l++){
                    if (k & (1 << l)){
                        shift += 1 << affect_args[mat_bit - 1 - l];
                    }
                }
                anc[now] += mat[i*mat_len + k] * vec[shift];
            }
        }

    }
    ''', 'matrix_dot_vector_double')


def matrix_dot_vector(
    vec: Union[np.ndarray, cp.ndarray],
    vec_bit: int,
    mat: Union[np.ndarray, cp.ndarray],
    mat_args: List[int],
    sync: bool = True
):
    # Matrix property
    mat_bit = np.int32(len(mat_args))
    mat_length = np.int32(2 ** mat_bit)
    assert vec_bit >= mat_bit, "Vector length should larger than matrix."

    if vec_bit == mat_bit:
        return dot(mat, vec, sync=sync)

    # GPU preparation
    task_number = 1 << (vec_bit - mat_bit)
    thread_per_block = min(256, task_number)
    block_num = task_number // thread_per_block

    # matrix args and sorted args
    for i in range(mat_bit):
        mat_args[i] = vec_bit - 1 - mat_args[i]

    sorted_mat_args = mat_args.copy()
    sorted_mat_args.sort()
    mat_args = cp.array(mat_args, dtype=np.int32)
    sorted_mat_args = cp.array(sorted_mat_args, dtype=np.int32)

    # Vector, Matrix preparation
    if isinstance(vec, np.ndarray):
        vec = cp.array(vec, dtype=vec.dtype)

    if isinstance(mat, np.ndarray):
        mat = cp.array(mat, dtype=mat.dtype)
    auxiliary_vec = cp.zeros_like(vec, dtype=vec.dtype)

    # Start GPU kernel function
    kernel_function = matrix_dot_vector_single_kernel if vec.dtype == np.complex64 else matrix_dot_vector_double_kernel
    kernel_function(
        (block_num,),
        (thread_per_block,),
        (mat, mat_bit, mat_length, vec, mat_args, sorted_mat_args, auxiliary_vec)
    )

    if sync:
        cp.cuda.Device().synchronize()

    return auxiliary_vec


kernel_funcs = list(locals().keys())
for name in kernel_funcs:
    if name.endswith("kernel"):
        locals()[name].compile()



from .data_structure import JobOperatorType, JobState, ResourceOp
from .dynamic_resource_manage import user_resource_op
from .file_manage import delete_job_folder, delete_user_folder, create_job_folder, create_user_folder
from .get_config import get_default_user_config



from enum import Enum


class JobOperatorType(Enum):
    delete = "DEL"
    user_delete = "UDL"


class ResourceOp(Enum):
    Allocation = "ALC"
    Release = "RES"


class JobState(Enum):
    Pending = "pending"
    Running = "running"
    Finish = "finish"

    # Un-normal State
    Error = "error"



import random
import numpy as np

from QuICT.core import Circuit
from QuICT.core.gate import *


class BenchmarkCircuitBuilder:
    """
    A class fetch QuICT benchmark circuits.

     Args:
        width(int): number of qubits
        size(int): number of gates
        random_params (bool, optional): whether random parameter or use default parameter. Defaults to True.
    """
    @staticmethod
    def parallelized_circuit_build(width: int, size: int, random_params: bool = True):
        typelist = [GateType.rz, GateType.cx]
        prob = [0.8, 0.2]

        gate_indexes = list(range(2))
        qubits_indexes = list(range(width))
        shuffle_qindexes = qubits_indexes[:]
        random.shuffle(shuffle_qindexes)

        cir = Circuit(width)
        while cir.size() < size:
            rand_type = np.random.choice(gate_indexes, p=prob)
            gate_type = typelist[rand_type]
            gate = GATE_TYPE_TO_CLASS[gate_type]()

            if random_params and gate.params:
                gate.pargs = list(np.random.uniform(0, 2 * np.pi, gate.params))

            gsize = gate.controls + gate.targets
            if gsize > len(shuffle_qindexes):
                continue

            gate & shuffle_qindexes[:gsize] | cir

            if gsize == len(shuffle_qindexes):
                shuffle_qindexes = qubits_indexes[:]
                random.shuffle(shuffle_qindexes)
            else:
                shuffle_qindexes = shuffle_qindexes[gsize:]

        return cir

    @staticmethod
    def serialized_circuit_build(width: int, size: int, random_params: bool = True):
        cir = Circuit(width)
        qubit_indexes = list(range(width))
        qubit = random.choice(qubit_indexes)
        qubit_indexes.remove(qubit)

        while cir.size() < size:
            qubit_new = random.choice(qubit_indexes)
            qubits_list = [qubit, qubit_new]
            random.shuffle(qubits_list)
            CX & (qubits_list) | cir

        return cir

    @staticmethod
    def entangled_circuit_build(width: int, size: int, random_params: bool = True):
        def _pattern1():
            cgate = CompositeGate()
            qubit_indexes = list(range(width))
            qubit_extra = []
            for _ in range(width):
                if len(qubit_indexes) > 1:
                    qubit_index = random.sample(qubit_indexes, 2)
                    CX & (qubit_index) | cgate
                    qubit_extra.append(qubit_index)
                    qubit_indexes = list(set(qubit_indexes) - set(qubit_index))
                elif len(qubit_indexes) == 1:
                    for i in range(len(qubit_extra)):
                        q_collect = random.choice(qubit_extra[i])
                        CX & ([qubit_indexes[0], q_collect]) | cgate
                    break
                else:
                    break

            return cgate

        def _pattern2():
            cgate = CompositeGate()
            qubit_indexes = list(range(width))
            result = [qubit_indexes[i:i + 2] for i in range(0, len(qubit_indexes), 2)]
            for i in range(len(result)):
                if len(result[i]) == 2:
                    CX & (result[i]) | cgate

            result = [qubit_indexes[i + 1:i + 3] for i in range(0, len(qubit_indexes), 2)]
            for i in range(len(result)):
                if len(result[i]) == 2:
                    CX & (result[i]) | cgate
                else:
                    break

            return cgate

        cir = Circuit(width)
        while cir.size() < size:
            cgate = random.choice([_pattern1(), _pattern2()])
            cgate | cir

        return cir

    @staticmethod
    def mediate_measure_circuit_build(width: int, size: int, random_params: bool = True):
        typelist = [GateType.rz, GateType.cx]
        prob = [0.8, 0.2]

        def flat_build():
            cgate = CompositeGate()
            qubits_indexes = list(range(width))
            random.shuffle(qubits_indexes)
            while len(qubits_indexes) > 0:
                if len(qubits_indexes) == 1:
                    Rz & qubits_indexes.pop() | cgate
                else:
                    gate_type = np.random.choice(typelist, p=prob)
                    gate = GATE_TYPE_TO_CLASS[gate_type]()
                    gate_size = gate.controls + gate.targets
                    gate & qubits_indexes[:gate_size] | cgate
                    qubits_indexes = qubits_indexes[gate_size:]

            return cgate

        cir = Circuit(width)
        flat_build() | cir
        cir.random_append(size - 2 * width, typelist, random_params, prob)
        flat_build() | cir

        idxes = random.sample(list(range(width, size - width)), k=width)
        for i in range(width):
            mgate = Measure & i
            cir.replace_gate(idxes[i], mgate)

        return cir



from .. import InstructionSet
from ..transform_rule import hrz_rule

from QuICT.core.gate import *

NamSet = InstructionSet(
    GateType.cx,
    [GateType.h, GateType.rz]
)
NamSet.register_one_qubit_rule(hrz_rule)



from typing import *
import numpy as np

from .utility import *


class BlockLDUDecompose:
    @classmethod
    def remapping_select(cls, mat_: np.ndarray) \
            -> List[int]:
        """
        Select some rows out of a square matrix to get a full-ranked sub-matrix.

        Args:
            mat_(np.ndarray): Boolean matrix to be decomposed

        Returns:
            List[int]: Row reordering args. The remapping ensures that if selected rows
                are in the upper part before selecting, their locations won't be changed.
        """
        mat = mat_.copy()
        n = mat.shape[0]
        s_size = n // 2
        selected = []
        unselected = []
        rk = 0
        for i in range(n):
            selected.append(i)
            sub_mat = mat[selected, :s_size]
            slct_rk = f2_rank(sub_mat)
            if slct_rk == len(selected):
                rk += 1
                if rk == s_size:
                    for j in range(i + 1, n):
                        unselected.append(j)
                    break
            else:
                unselected.append(selected.pop())

        selected_part2 = []
        swap_flag = [True for _ in range(s_size)]
        for i in range(s_size):
            if selected[i] < s_size:
                swap_flag[selected[i]] = False
            else:
                selected_part2.append(selected[i])

        remapping = [i for i in range(n)]

        cur_part2_ptr = 0
        for i in range(s_size):
            if swap_flag[i]:
                x = i
                y = selected_part2[cur_part2_ptr]
                remapping[x], remapping[y] = remapping[y], remapping[x]
                cur_part2_ptr += 1

        return remapping

    @classmethod
    def run(cls, mat_: np.ndarray) \
            -> Tuple[np.ndarray, np.ndarray, np.ndarray, List[int]]:
        """
        Block LDU decomposition of a boolean matrix.

        Args:
            mat_(np.ndarray): Boolean matrix to be decomposed

        Returns:
            Tuple[np.ndarray, np.ndarray, np.ndarray, List[int]]: Decomposed L, D, U and row remapping.
        """

        n = mat_.shape[0]
        if n <= 1:
            raise Exception("LDU decomposition is designed for matrix whose size >= 2.")

        remapping = cls.remapping_select(mat_)
        mat: np.ndarray = mat_.copy()[remapping]

        s_size = n // 2
        t_size = n - s_size

        a = mat[:s_size, :s_size]
        b = mat[:s_size, s_size:]
        c = mat[s_size:, :s_size]
        d = mat[s_size:, s_size:]

        ai = f2_inverse(a)
        c_ai = f2_matmul(c, ai)
        c_ai_b = f2_matmul(c_ai, b)
        ai_b = f2_matmul(ai, b)

        lower = np.zeros(shape=(n, n), dtype=bool)
        diagonal = np.zeros(shape=(n, n), dtype=bool)
        upper = np.zeros(shape=(n, n), dtype=bool)

        # lower
        lower[:s_size, :s_size] = np.eye(s_size, dtype=bool)
        lower[s_size:, s_size:] = np.eye(t_size, dtype=bool)
        lower[s_size:, :s_size] = c_ai

        # diagonal
        diagonal[:s_size, :s_size] = a
        diagonal[s_size:, s_size:] = d ^ c_ai_b

        # upper
        upper[:s_size, :s_size] = np.eye(s_size, dtype=bool)
        upper[s_size:, s_size:] = np.eye(t_size, dtype=bool)
        upper[:s_size, s_size:] = ai_b

        return lower, diagonal, upper, remapping



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2021/3/20 4:22 下午
# @Author  : Han Yu
# @File    : __init__.py

from .instruction_set import InstructionSet
from .gate_transform import GateTransform
from .special_set import *
from .transform_rule import *



import numpy as np

from ._operator import Operator


class NoiseGate(Operator):
    """
    The quantum gate with noise error.

    Args:
        gate (BasicGate): The quantum gate.
        error (QuantumNoiseError): The noise error.
    """
    @property
    def noise_matrix(self) -> list:
        """ The noised gate matrix. """
        return self._noise_matrix

    @property
    def qasm_name(self) -> str:
        return "noise"

    @property
    def type(self) -> str:
        return "noise"

    def __init__(self, noise_matrix, args_num: int):
        super().__init__(args_num)
        self._noise_matrix = noise_matrix
        self._precision = self._noise_matrix[0].dtype

    def convert_precision(self):
        self._precision = np.complex128 if self._precision == np.complex64 else np.complex64
        for noise_matrix in self._noise_matrix:
            noise_matrix = noise_matrix.astype(self._precision)



import copy
import math
from random import random
from typing import Tuple, Union

import torch

from QuICT.core import *
from QuICT.core.gate import *
from QuICT.qcda.mapping.ai.data_def import State, StateSlim, TrainConfig
from QuICT.qcda.mapping.ai.net.nn_mapping import NnMapping


class Agent:
    def __init__(
        self,
        config: TrainConfig,
    ) -> None:
        # Copy values in.
        self.config = config

        # Exploration related.
        self.explore_step = 0
        self.state = None  # Reset during training

        self.mapped_circ = CompositeGate()
        self._last_exec_cnt = 0
        self._last_action = None

        # Initialize policy & target network
        assert config.topo is not None
        self._qubit_num = config.topo.qubit_number

    def reset_explore_state(self):
        self.state = self.config.factory.get_one()
        # Do we need to reset counter?
        # self.explore_step = 0

    def _select_action(
        self,
        policy_net: NnMapping,
        policy_net_device: str,
    ):
        config = self.config
        a = config.action_num

        # Choose an action based on policy_net
        data = self.state.to_slim().to_nn_data()
        data = StateSlim.batch_from_list([data], policy_net_device)
        # data = self.state.circ_layered_matrices.to(policy_net_device)
        q_vec = policy_net(data.x, data.edge_index, data.batch).detach().cpu()
        q_vec = q_vec.view(a)  # [a]
        if self._last_action is not None:
            bad_action = self._last_action
            if bad_action not in config.action_id_by_swap:
                bad_action = bad_action[1], bad_action[0]
            bad_id = config.action_id_by_swap[bad_action]
            q_vec[bad_id] = 1e-12

        # Query action swap using action id
        idx = int(torch.argmax(q_vec))
        action = config.swap_by_action_id[idx]
        return action

    def select_action(
        self,
        policy_net: NnMapping,
        policy_net_device: str,
        epsilon_random: bool = False,
    ) -> Tuple[int, int]:
        """Select action based on given policy net.

        Args:
            policy_net (GnnMapping): A policy network which gives an action as output.
            policy_net_device (str): Policy net model device.
            epsilon_random (bool, optional): Whether to use epsilon-greedy strategy.
                If set True, there will be epsilon probability that this methods randomly select a viable action.
                Defaults to False.

        Returns:
            Tuple[int, int]: Selected action.
        """
        config = self.config
        eps_threshold = config.epsilon_end + (
            config.epsilon_start - config.epsilon_end
        ) * math.exp(-1.0 * self.explore_step / config.epsilon_decay)
        if epsilon_random and random() <= eps_threshold:
            return self.state.biased_random_swap(exclude=(self._last_action,))
        else:
            return self._select_action(
                policy_net=policy_net,
                policy_net_device=policy_net_device,
            )

    def count_gate_num(self) -> int:
        return self.state.circ_info.count_gate()

    def take_action(
        self, action: Tuple[int, int]
    ) -> Tuple[StateSlim, Union[StateSlim, None], float, bool]:
        """Take given action on trainer's state.

        Args:
            action (Tuple[int, int]): Swap gate used on current topology.

        Returns:
            Tuple[StateSlim, Union[StateSlim, None], float, bool]: Tuple of
                (Previous State, Next State, Reward, Terminated).
        """
        assert self.state is not None

        config = self.config
        qubit_number = config.topo.qubit_number

        prev_state = self.state.to_slim()

        self.explore_step += 1

        u, v = action
        if u == -1 and v == -1:
            next_state = prev_state
            reward = -scale * 1000
            terminated = False
            return prev_state, next_state, reward, terminated

        topo_dist = self.state.layout_info.topo_dist
        scale = config.reward_scale
        if topo_dist[u][v] < 0.01:  # not connected
            assert False

        with self.mapped_circ:
            Swap & [u, v]

        next_logic2phy = self.state.logic2phy
        next_phy2logic = self.state.phy2logic

        # Current (u, v) are physical qubit labels
        next_phy2logic[u], next_phy2logic[v] = next_phy2logic[v], next_phy2logic[u]
        lu, lv = next_phy2logic[u], next_phy2logic[v]
        next_logic2phy[lu], next_logic2phy[lv] = next_logic2phy[lv], next_logic2phy[lu]

        self.state.logic2phy = next_logic2phy
        self.state.phy2logic = next_phy2logic

        # next_circ_state = self.state.circ_info.copy()
        action_penalty = -scale / qubit_number
        reward = action_penalty
        # Execute as many as possible
        cnt = self.state.eager_exec(
            physical_circ=self.mapped_circ,
        )
        if action == self._last_action:
            reward += -1 * scale
        # If no gate is executed, avoid the same selection next time.
        self._last_action = action if cnt == 0 else None
        self._last_exec_cnt = cnt
        reward += cnt * scale

        terminated = self.state.circ_info.count_gate() == 0

        next_state = None if terminated else self.state.to_slim()

        return prev_state, next_state, reward, terminated



from itertools import chain
from math import log2

from QuICT.core import *
from QuICT.core.gate import *


class PhasePolynomial:
    """
    Phase polynomial representation of a quantum circuit that only contains
    X, CX and Rz gates.
    """
    def __init__(self, gates):
        """
        Create a phase polynomial from given circuit. The circuit can only contain
        Controlled-X and Rz gates.

        Args:
            gates(DAG): Circuit represented by this polynomial
        """

        self.phases = {}
        self.gates = []
        self.size = gates.width()
        self._build_poly(gates)

    def _build_poly(self, gates):
        monomials = {}
        for gate_ in gates:
            gate_: BasicGate
            for qubit_ in chain(gate_.cargs, gate_.targs):
                if qubit_ not in monomials:
                    monomials[qubit_] = 1 << (qubit_ + 1)

            if gate_.qasm_name == 'cx':
                monomials[gate_.targ] = monomials[gate_.targ] ^ monomials[gate_.carg]
                self.gates.append(gate_)
            elif gate_.qasm_name == 'x':
                monomials[gate_.targ] = monomials[gate_.targ] ^ 1
                self.gates.append(gate_)
            elif gate_.qasm_name == 'rz':
                sign = -1 if monomials[gate_.targ] & 1 else 1
                mono = (monomials[gate_.targ] >> 1)
                self.phases[mono] = sign * gate_.parg + (self.phases[mono] if mono in self.phases else 0)
            else:
                raise Exception("PhasePolynomial only accepts cx, x and rz gates.")

    def get_circuit(self, epsilon=1e-10):
        """
        Generate a circuit of minimal size that implements the phase polynomial

        Returns:
            Circuit: Circuit equivalent to this polynomial
        """
        max_monomial = max(self.phases.keys()) if len(self.phases.keys()) > 0 else 1
        circ = Circuit(self.size)
        visited = set()
        for qubit_ in range(int(np.ceil(log2(max_monomial + 1))) + 1):
            if (1 << qubit_) in self.phases:
                if abs(self.phases[1 << qubit_]) > epsilon:
                    Rz(self.phases[1 << qubit_]) | circ(qubit_)
                visited.add(1 << qubit_)

        monomials = {}
        for gate_ in self.gates:
            gate_: BasicGate
            for qubit_ in chain(gate_.cargs, gate_.targs):
                if qubit_ not in monomials:
                    monomials[qubit_] = 1 << (qubit_ + 1)

            if gate_.qasm_name == 'cx':
                monomials[gate_.targ] = monomials[gate_.targ] ^ monomials[gate_.carg]
            elif gate_.qasm_name == 'x':
                monomials[gate_.targ] = monomials[gate_.targ] ^ 1

            gate_.copy() | circ(list(chain(gate_.cargs, gate_.targs)))
            cur_phase = monomials[gate_.targ] >> 1
            cur_sign = -1 if (monomials[gate_.targ] & 1) else 1

            if cur_phase in self.phases and cur_phase not in visited:
                if abs(self.phases[cur_phase]) > epsilon:
                    Rz(cur_sign * self.phases[cur_phase]) | circ(gate_.targ)
                visited.add(cur_phase)

        return circ



from QuICT.qcda.synthesis.arithmetic.vbe import VBEAdder, VBEAdderMod
from QuICT.qcda.synthesis.arithmetic.bea import BEAAdder, BEAAdderMod
from QuICT.qcda.synthesis.arithmetic.hrs import HRSAdder, HRSAdderMod
from QuICT.qcda.synthesis.arithmetic.tmvh import RippleCarryAdder



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2020/12/27 5:03 下午
# @Author  : Han Yu
# @File    : topology.py

from __future__ import annotations
from itertools import combinations

import json
import warnings
from typing import Any, Dict, List, Tuple


class LayoutEdge:
    """Implement a physical connection between physical qubits

    Attributes:
        u(int): Node u of edge
        v(int): Node v of edge
        directional(bool): Whether the layout is directional
        error_rate(float): Error_rate between u and v, default 1.0
    """

    @property
    def u(self) -> int:
        return self._u

    @u.setter
    def u(self, u):
        warnings.warn("In general, the LayoutEdge shouldn't be written.")
        self._u = u

    @property
    def v(self) -> int:
        return self._v

    @v.setter
    def v(self, v):
        warnings.warn("In general, the LayoutEdge shouldn't be written.")
        self._v = v

    @property
    def error_rate(self) -> float:
        return self._error_rate

    @error_rate.setter
    def error_rate(self, error_rate):
        self._error_rate = error_rate

    @property
    def directional(self):
        return self._directional

    @directional.setter
    def directional(self, value: bool):
        self._directional = value

    def to_dict(self) -> Dict[str, Any]:
        data = {}
        data["u"] = self._u
        data["v"] = self._v
        data["error_rate"] = self._error_rate
        data["directional"] = self._directional
        return data

    def __init__(self, u: int, v: int, directional: bool, error_rate: float):
        self._u = u
        self._v = v
        self._error_rate = error_rate
        self._directional = directional

    def __str__(self):
        dir = ' <-> ' if not self._directional else '-->'
        return f"{self._u} {dir} {self._v}, with error rate {self._error_rate}"


class Layout:
    """Implement a topology in a physical device

    Attributes:
        name(string): the name of the topology
        edge_list(list<LayoutEdge>): the edge in layout
        qubit_number(int): the number of qubits
    """

    DIRECTIONAL_DEFAULT = False
    ERROR_RATE_DEFAULT = 1.0

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def edge_list(self) -> List[LayoutEdge]:
        return list(self._edges.values())

    def __iter__(self):
        yield from self._edges.values()

    @property
    def qubit_number(self):
        return self._qubit_number

    def __init__(self, qubit_number: int, name: str = "unknown"):
        self._qubit_number = qubit_number
        self._name = name
        self._edges: Dict[Tuple[int, int], LayoutEdge] = {}
        self._directionalized = None

    def __str__(self):
        layout_str = f"{self._name} with {self._qubit_number} qubits."
        for edge in self.edge_list:
            layout_str += f"\n{edge}"

        return layout_str

    def out_edges(self, begin_point: int) -> List[LayoutEdge]:
        """edges begin from begin_point

        Args:
            begin_point(int): the index of begin node

        Return:
            list<LayoutEdge>: edges begin from begin_point
        """
        prior_list = []
        for edge in self.edge_list:
            if edge.u == begin_point:
                prior_list.append(edge)
        return prior_list

    @property
    def directionalized(self) -> Layout:
        """Return a copy of current layout with all un-directional edges
        replaced with 2 reversed directional edges.
        """
        layout = Layout(self._qubit_number, self._name)
        for edge in self:
            if edge.directional:
                layout.add_edge(edge)
            else:
                layout.add_edge(
                    u=edge.u, v=edge.v, directional=True, error_rate=edge.error_rate
                )
                layout.add_edge(
                    u=edge.v, v=edge.u, directional=True, error_rate=edge.error_rate
                )
        return layout

    def add_edge(
        self,
        u,
        v,
        directional=DIRECTIONAL_DEFAULT,
        error_rate=ERROR_RATE_DEFAULT,
    ):
        """add an edge in the layout

        Args:
            u(int): Edge endpoint u
            v(int): Edge endpoint v
            directional(bool): Whether the edge is directional
            error_rate(float): Error rate, default 1.0
        """
        assert u != v, "Endpoints cannot be the same"
        edge = LayoutEdge(u, v, directional, error_rate)
        key = (u, v) if directional or u < v else (v, u)
        self._edges[key] = edge
        # Reset cache
        self._directionalized = None

    def check_edge(self, u, v):
        """Check whether layout contain u->v

        Args:
            u(int): the edge endpoint u
            v(int): the edge endpoint v
        Return:
            bool: whether layout contain u->v
        """
        return ((u, v) in self._edges) or (
            (v, u) in self._edges and not self._edges[(v, u)].directional
        )

    def to_json(self) -> str:
        """Serialize current layout into json string."""
        data = {}
        data["name"] = self._name
        data["qubit_number"] = self._qubit_number
        edges = [edge.to_dict() for edge in self]
        data["edges"] = edges
        return json.dumps(data)

    def write_file(self, directory="./"):
        """Write layout into file.

        Args:
            directory(str): Directory to store layout file in, default "./"
        """
        with open(f"{directory}{self.name}.layout", "w") as f:
            f.write(self.to_json())

    @classmethod
    def from_json(cls, json_str: str) -> Layout:
        json_obj = json.loads(json_str)
        name = json_obj["name"]
        qubit_number = json_obj["qubit_number"]
        layout = Layout(qubit_number, name)
        edges = json_obj["edges"]
        for edge in edges:
            u = edge["u"]
            v = edge["v"]
            directional = (
                edge["directional"]
                if "directional" in edge
                else cls.DIRECTIONAL_DEFAULT
            )
            error_rate = (
                edge["error_rate"] if "error_rate" in edge else cls.ERROR_RATE_DEFAULT
            )
            layout.add_edge(u=u, v=v, directional=directional, error_rate=error_rate)
        return layout

    @classmethod
    def load_file(cls, file_path: str) -> Layout:
        """Load layout from file.
        Args:
            file_path(str): Path of layout file.
        Return:
            Layout: Layout parsed from file.
        """
        with open(file_path) as f:
            return cls.from_json(f.read())

    def get_sublayout_edges(self, qubits: list) -> list:
        connected_qubits = [qubits[0]]
        connected_edges = []
        for qubit in qubits:
            edges = self.out_edges(qubit)
            for edge in edges:
                if edge.v in qubits:
                    if edge.v not in connected_qubits:
                        connected_qubits.append(edge.v)

                    target_edge = set([edge.u, edge.v])
                    if target_edge not in connected_edges:
                        connected_edges.append(target_edge)

        if len(connected_qubits) == len(qubits):
            return connected_edges
        else:
            return []

    def sub_layout(self, qubits_number: int):
        """ Get partial layout. Only working for undirectional layout.

        Args:
            qubits_number (int): The number of qubits for sub-layout

        Returns:
            Layout: The sub-layout
        """
        assert qubits_number < self.qubit_number

        qubits_list = list(range(self.qubit_number))
        all_combined = list(combinations(qubits_list, qubits_number))
        num_edges_subl, edges_subl = [], []
        for q_comb in all_combined:
            edges_sl = self.get_sublayout_edges(q_comb)
            num_edges_subl.append(len(edges_sl))
            edges_subl.append(edges_sl)

        max_value = max(num_edges_subl)
        if max_value <= 0:
            raise ValueError("Failure to find sub-layout.")

        max_idx = num_edges_subl.index(max_value)
        sub_layout = self._get_layout(all_combined[max_idx], edges_subl[max_idx])

        return sub_layout

    def _get_layout(self, related_qubits: list, edges: list) -> Layout:
        sub_layout = Layout(len(related_qubits))
        index_mapping = {}
        for idx, q in enumerate(related_qubits):
            index_mapping[q] = idx

        for u, v in edges:
            sub_layout.add_edge(index_mapping[u], index_mapping[v], directional=False)

        return sub_layout



# Author  : Peng Sirui

"""
The (2n+3)-qubit circuit used in the Shor algorithm is designed by \
Stephane Beauregard in "Circuit for Shor's algorithm using 2n+3 qubits"\

Without one control-bit trick.
"""

from math import gcd, pi
import numpy as np
from fractions import Fraction
from functools import reduce

from QuICT.core import Circuit
from QuICT.core.gate import *
from QuICT.qcda.synthesis.arithmetic.bea import *
from QuICT.simulation.state_vector import CircuitSimulator
from .utility import *

from QuICT.tools import Logger
from QuICT.tools.exception.core import *

logger = Logger("BEA")


def construct_circuit(a: int, N: int, eps: float = 1 / 10):
    # phase estimation procedure
    n = int(np.ceil(np.log2(N + 1)))
    t = int(2 * n + 1 + np.ceil(np.log(2 + 1 / (2 * eps))))
    logger.info(f"\torder_finding circuit construction: circuit: n = {n} t = {t}")

    circuit = Circuit(2 * n + 2 + t)
    b_reg = list(range(n + 1))  # n+1
    x_reg = list(range(n + 1, 2 * n + 1))  # n
    trickbits = list(range(2 * n + 1, 2 * n + 1 + t))  # t
    qreg_low = [2 * n + 1 + t]  # 1
    X | circuit(x_reg[n - 1])

    for idx in trickbits:
        H | circuit(idx)
    for k in range(t):
        gate_pow = pow(a, 1 << (t - 1 - k), N)  # CUa^{2^{t-1-k}}
        BEACUa.execute(n, gate_pow, N) | circuit(
            b_reg + x_reg + [trickbits[k]] + qreg_low
        )
    for k in range(len(trickbits) // 2):
        Swap | circuit([trickbits[k], trickbits[len(trickbits) - 1 - k]])
    IQFT.build_gate(len(trickbits)) | circuit(trickbits)
    for k in range(len(trickbits) // 2):
        Swap | circuit([trickbits[k], trickbits[len(trickbits) - 1 - k]])
    for idx in b_reg + trickbits + qreg_low:
        Measure | circuit(idx)
    return circuit, trickbits[::-1]  # for int(circuit[trickbits]) convenience


def order_finding(a: int, N: int, eps: float = 1 / 10, simulator=CircuitSimulator()):
    circuit, trickbits = construct_circuit(a, N, eps)
    simulator.run(circuit)
    t = len(trickbits)

    # continued fraction procedure
    phi_ = int(circuit[trickbits]) / (1 << t)
    logger.info(f"\tphi~ (approximately s/r) in decimal form is {phi_}")
    r = Fraction(phi_).limit_denominator(N - 1).denominator
    logger.info(f"\tclose fraction form: {Fraction(phi_).limit_denominator(N - 1)}")
    return r


MAX_ROUND = 3


def reinforced_order_finding(
    a: int, N: int, eps: float = 1 / 10, simulator=CircuitSimulator()
):
    circuit, trickbits = construct_circuit(a, N, eps)
    t = len(trickbits)
    # continued fraction procedure (repetition)
    r_list = []
    i = 0
    while i < MAX_ROUND:
        i += 1
        simulator.run(circuit)
        phi_ = int(circuit[trickbits]) / (1 << t)
        logger.info(
            f"\tclose fraction form (repetition {i}): {Fraction(phi_).limit_denominator(N - 1)}"
        )
        r = Fraction(phi_).limit_denominator(N - 1).denominator
        if r != 0 and (a ** r) % N == 1:
            logger.info("\tsuccess!")
            r_list.append(r)
    if len(r_list) == 0:
        return 0
    else:
        return reduce(lambda x, y: (x * y) // gcd(x, y), r_list)



import os
import shutil

from QuICT.tools import Logger
from QuICT.tools.cli.utils import JobValidation
from QuICT.tools.circuit_library import CircuitLib


logger = Logger("CLI_Circuit_Management")


default_customed_circuit_folder = os.path.join(
    os.path.dirname(__file__),
    "..",
    "circuit"
)


def get_random_circuit(
    qubits: list,
    max_size: int,
    max_depth: int,
    instruction_set: str = "nam",
    output_path: str = '.'
):
    """ Generate the circuit with give parameters and write circuit's qasm into output path.

    Args:
        qubits (list): The list of qubits number.
        size (list): The list of circuit's gate number.
        random_param (bool): whether using random parameters for all quantum gates with parameters.
        instruction_set (str, optional): The given instruction sets. Defaults to "random".
        output_path (str, optional): The output folder. Defaults to current work dir.
    """
    circuit_library = CircuitLib(output_type="file", output_path=output_path)
    circuit_library.get_random_circuit(instruction_set, qubits, max_size, max_depth)


def get_algorithm_circuit(alg: str, qubits: list, output_path: str = "."):
    """ Get the algorithm circuit and write its qasm into output path. """
    circuit_library = CircuitLib(output_type="file", output_path=output_path)
    circuit_library.get_algorithm_circuit(alg, qubits_interval=qubits)


def store_quantum_circuit(name: str, file: str):
    """ Save the given quantum circuit into Quantum Circuit Library. """
    get_folder_name = os.listdir(default_customed_circuit_folder)
    if not name.endswith(".qasm"):
        name += ".qasm"

    if name in get_folder_name:
        logger.warn("Repeat circuits name.")
        return

    # qasm file validation
    _ = JobValidation.get_circuit_info(file)
    shutil.copy(file, f"{default_customed_circuit_folder}/{name}")
    logger.info(f"Successfully add circuit {name} into Circuit Library.")


def delete_quantum_circuit(name: str):
    """ Delete the customed quantum circuit in Circuit Library. """
    get_folder_name = os.listdir(default_customed_circuit_folder)
    if not name.endswith(".qasm"):
        name += ".qasm"

    if name not in get_folder_name:
        logger.warn("No target circuit in Circuit Library.")
        return

    os.remove(f"{default_customed_circuit_folder}/{name}")
    logger.info(f"Successfully remove circuit {name} from Circuit Library.")


def list_quantum_circuit():
    """ List all customed quantum circuit in Circuit Library. """
    circuit_name = os.listdir(default_customed_circuit_folder)
    logger.info(f"There are {len(circuit_name)} in the Circuit Library.")
    if len(circuit_name) > 0:
        logger.info("circuit names: " + ", ".join([cir[:-5] for cir in circuit_name]))



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2022/1/14 9:36 下午
# @Author  : Han Yu, Li Kaiqi
# @File    : composite_gate.py
from typing import Union
import numpy as np

from QuICT.core.qubit import Qureg, Qubit
from QuICT.core.gate import BasicGate
from QuICT.core.utils import (
    CircuitBased,
    CGATE_LIST,
    unique_id_generator
)
from QuICT.tools.exception.core import ValueError, CompositeGateAppendError, TypeError, GateQubitAssignedError


class CompositeGate(CircuitBased):
    """ Implement a group of gate

    Attributes:
        name (str): the name of the composite gate
        gates (list<BasicGate>): gates within this composite gate
    """
    @property
    def checkpoint(self):
        return self._check_point

    def __init__(self, name: str = None, gates: list = None):
        if name is None:
            name = "composite_gate_" + unique_id_generator()

        super().__init__(name)
        self._check_point = None        # required checkpoint
        self._min_qubit = np.inf
        self._max_qubit = 0

        if gates is not None:
            self.extend(gates)

    ####################################################################
    ############          CompositeGate Context             ############
    ####################################################################
    def __enter__(self):
        global CGATE_LIST
        CGATE_LIST.append(self)

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        # print(f"{exc_type}: {exc_value}")
        if exc_type is not None:
            raise Exception(exc_value)

        global CGATE_LIST
        CGATE_LIST.remove(self)

        return True

    ####################################################################
    ############        CompositeGate Qureg Mapping         ############
    ####################################################################
    def width(self):
        """ the number of qubits applied by gates

        Returns:
            int: the number of qubits applied by gates
        """
        return self._max_qubit

    def __call__(self, indexes: Union[list, int]):
        if isinstance(indexes, int):
            indexes = [indexes]

        self._update_qubit_limit(indexes)
        self._pointer = indexes
        return self

    def clean(self):
        self._gates = []
        self._min_qubit, self._max_qubit = np.inf, 0
        self._pointer = None

    def __and__(self, targets: Union[int, list, Qubit, Qureg]):
        """ assign qubits or indexes for given gates

        Args:
            targets ([int/qubit/list[int]/qureg]): qubit describe
        """
        if isinstance(targets, int):
            targets = [targets]

        if isinstance(targets, Qubit):
            targets = Qureg(targets)

        if len(targets) != self._max_qubit:
            raise ValueError("CompositeGate.&:len(targets)", f"less than {self._max_qubit}", len(targets))

        self._mapping(targets)
        if CGATE_LIST:
            CGATE_LIST[-1].extend(self.gates)

    def _mapping(self, targets: Qureg):
        """ remapping the gates' affectArgs

        Args:
            targets(Qureg/List): the related qubits
        """
        for gate in self._gates:
            args_index = gate.cargs + gate.targs
            if isinstance(targets, Qureg):
                target_qureg = targets(args_index)
                gate.assigned_qubits = target_qureg
                gate.update_name(target_qureg[0].id)
            else:
                gate.cargs = [targets[carg] for carg in gate.cargs]
                gate.targs = [targets[targ] for targ in gate.targs]

    def _update_qubit_limit(self, indexes: list):
        for idx in indexes:
            assert idx >= 0 and isinstance(idx, int)
            if idx >= self._max_qubit:
                self._max_qubit = idx + 1

            if idx < self._min_qubit:
                self._min_qubit = idx

    ####################################################################
    ############            CompositeGate Build             ############
    ####################################################################
    def __or__(self, targets):
        """ deal the operator '|'

        Use the syntax "gateSet | circuit", "gateSet | gateSet"
        to add the gate of gateSet into the circuit

        Note that the order of qubits is that control bits first
        and target bits followed.

        Args:
            targets: the targets the gate acts on, it can have following form,
                1) Circuit
                2) CompositeGate
        Raise:
            TypeError: the type of other is wrong
        """
        try:
            targets.extend(self)
        except Exception as e:
            raise CompositeGateAppendError(f"Failure to append current CompositeGate, due to {e}.")

    def __xor__(self, targets):
        """deal the operator '^'

        Use the syntax "gateSet ^ circuit", "gateSet ^ gateSet"
        to add the gate of gateSet's inverse into the circuit

        Note that the order of qubits is that control bits first
        and target bits followed.

        Args:
            targets: the targets the gate acts on, it can have following form,
                1) Circuit
        Raise:
            TypeError: the type of other is wrong
        """
        try:
            targets.extend(self.inverse().gates)
        except Exception as e:
            raise CompositeGateAppendError(f"Failure to append the inverse of current CompositeGate, due to {e}.")

    def __getitem__(self, item):
        """ get gates from this composite gate

        Args:
            item(int/slice): slice passed in.

        Return:
            [BasicGates]: the gates
        """
        return self._gates[item]

    def extend(self, gates: list):
        for gate in gates:
            self.append(gate, is_extend=True)

        self._pointer = None

    def append(self, gate, is_extend: bool = False, insert_idx: int = -1):
        from QuICT.core.operator import CheckPointChild, Operator

        if isinstance(gate, BasicGate):
            self._append_gate(gate, insert_idx)
            if not is_extend:
                self._pointer = None

            # Update gate type dict
            if gate.type in self._gate_type.keys():
                self._gate_type[gate.type] += 1
            else:
                self._gate_type[gate.type] = 1
        elif isinstance(gate, CheckPointChild):
            self._check_point = gate
        else:
            assert isinstance(gate, Operator), TypeError("CompositeGate.append", "BasicGate/Operator", type(gate))
            if insert_idx == -1:
                self._gates.append(gate)
            else:
                self._gates.insert(insert_idx, gate)

    def left_extend(self, gates: list):
        for idx, gate in enumerate(gates):
            self.append(gate, is_extend=True, insert_idx=idx)

        self._pointer = None

    def _append_gate(self, gate, insert_idx: int = -1):
        gate = gate.copy()

        if self._pointer is not None:
            qubit_index = self._pointer[:]
            gate_args = gate.controls + gate.targets
            if len(self._pointer) > gate_args:
                gate.cargs = [qubit_index[carg] for carg in gate.cargs]
                gate.targs = [qubit_index[targ] for targ in gate.targs]
            elif len(self._pointer) == gate_args:
                gate.cargs = qubit_index[:gate.controls]
                gate.targs = qubit_index[gate.controls:]
            else:
                raise GateQubitAssignedError(f"{gate.type} need {gate_args} indexes, but given {len(self._pointer)}")
        else:
            qubit_index = gate.cargs + gate.targs
            if not qubit_index:
                raise GateQubitAssignedError(f"{gate.type} need qubit indexes to add into Composite Gate.")

            self._update_qubit_limit(qubit_index)

        if insert_idx == -1:
            self._gates.append(gate)
        else:
            self._gates.insert(insert_idx, gate)

    ####################################################################
    ############            CompositeGate Utils             ############
    ####################################################################
    def inverse(self):
        """ the inverse of CompositeGate

        Returns:
            CompositeGate: the inverse of the gateSet
        """
        inverse_cgate = CompositeGate()
        inverse_gates = [gate.inverse() for gate in self._gates[::-1]]
        inverse_cgate.extend(inverse_gates)

        return inverse_cgate

    def matrix(self, device: str = "CPU", local: bool = False) -> np.ndarray:
        """ matrix of these gates

        Args:
            device (str, optional): The device type for generate circuit's matrix, one of [CPU, GPU]. Defaults to "CPU".
            local: whether regards the min_qubit as the 0's qubit

        Returns:
            np.ndarray: the matrix of the gates
        """
        if local and isinstance(self._min_qubit, int):
            min_value = self._min_qubit
        else:
            min_value = 0

        return super().matrix(device, min_value)



import os
import sqlite3
from datetime import datetime


def sql_locked(function):
    def decorator(self, *args, **kwargs):
        self._connect.execute('BEGIN EXCLUSIVE')
        result = function(self, *args, **kwargs)
        self._connect.commit()

        return result

    return decorator


class SQLManager:
    """ Using SQL database to store running job information in local mode, and
    login information for remote mode.
    """
    def __init__(self):
        file_path = os.path.join(os.path.dirname(__file__), "../../../../lib")
        self._connect = sqlite3.connect(f"{file_path}/CLI_runtime_data.db")
        self._connect.isolation_level = "EXCLUSIVE"
        self._cursor = self._connect.cursor()

        # Initial Database when first run
        try:
            self._create_table()
        except:
            pass

    def __del__(self):
        self._cursor.close()
        self._connect.close()

    def clean(self):
        """ Clean the table data in current DB. """
        self._cursor.execute("DROP TABLE JOB_INFO")
        self._cursor.execute("DROP TABLE USER_LOGIN_INFO")

    def _create_table(self):
        # Temp add here, delete after create datebase
        self._cursor.execute('CREATE TABLE JOB_INFO(NAME text unique, STATUS text, PID INT)')
        self._cursor.execute(
            'CREATE TABLE USER_LOGIN_INFO(' +
            'ID INT, NAME text unique, PASSWORD text, LOGINTIME text)'
        )

    ####################################################################
    ############               User DB Function             ############
    ####################################################################

    def user_login(self, username: str, password: str):
        """ Update user login information for CLI Remote Mode.

        Args:
            username(str): The username
            password(str): The encrypted password
        """
        login_time = datetime.now().strftime('%Y-%m-%d %H:%M:%S')

        if self.get_login_info() is None:
            self._cursor.execute(
                "INSERT INTO USER_LOGIN_INFO(ID, NAME, PASSWORD, LOGINTIME)" +
                f"VALUES (1, \'{username}\', \'{password}\', \'{login_time}\')"
            )
        else:
            self._cursor.execute(
                "UPDATE USER_LOGIN_INFO " +
                f"SET NAME = \'{username}\', PASSWORD = \'{password}\', LOGINTIME =  \'{login_time}\' " +
                "WHERE ID = \'1\'"
            )

        self._connect.commit()

    def user_logout(self):
        """ Clean current login information in database. """
        self._cursor.execute("DELETE FROM USER_LOGIN_INFO WHERE ID = \'3\'")
        self._connect.commit()

    def login_validation(self) -> bool:
        """ validation the login information including successful login user and expired time. """
        user_info = self.get_login_info()
        if user_info is None:
            return False, "Please login first."

        # login time check
        current_time = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        login_time = user_info[3]
        time_diff = datetime.strptime(current_time, '%Y-%m-%d %H:%M:%S') - \
            datetime.strptime(login_time, '%Y-%m-%d %H:%M:%S')
        if time_diff.seconds > 3600:
            return False, "Please login again. The last login is expired."

        return True, None

    def get_login_info(self) -> tuple:
        """ return login information

        Returns:
            tuple: (id, username, password, logintime)
        """
        self._cursor.execute('SELECT * FROM USER_LOGIN_INFO')
        return self._cursor.fetchone()

    ####################################################################
    ############               Job DB Function              ############
    ####################################################################

    def job_validation(self, job_name: str) -> bool:
        """ Validation Job is existed in current DB. """
        job_info = self._cursor.execute(f"SELECT * FROM JOB_INFO WHERE NAME=\'{job_name}\'")
        job_info = self._cursor.fetchone()

        return job_info is not None

    @sql_locked
    def add_job(self, job_info: dict):
        """ Add new job into DB. """
        name = job_info['name']     # Job name
        status = job_info['status']
        pid = job_info['pid']

        # Add job information into JOB_INFO
        self._cursor.execute(f"INSERT INTO JOB_INFO VALUES (\'{name}\', \'{status}\', \'{pid}\')")

    @sql_locked
    def get_job_status(self, job_name: str):
        """ Return the given job's states. """
        self._cursor.execute(f'SELECT STATUS FROM JOB_INFO WHERE NAME=\'{job_name}\'')
        job_status = self._cursor.fetchone()

        return job_status[0]

    @sql_locked
    def get_job_pid(self, job_name: str):
        """ Return the given job's pid. """
        self._cursor.execute(f'SELECT PID FROM JOB_INFO WHERE NAME=\'{job_name}\'')
        job_status = self._cursor.fetchone()

        return job_status[0]

    @sql_locked
    def change_job_status(self, job_name: str, job_status: str):
        """ Change the given job's state. """
        self._cursor.execute(f"UPDATE JOB_INFO SET STATUS = \'{job_status}\' WHERE NAME = \'{job_name}\'")

    @sql_locked
    def delete_job(self, job_name: str):
        """ Delete the given job from DB. """
        try:
            # DELETE FROM USER_PASS_MAPPING
            self._cursor.execute(f"DELETE FROM JOB_INFO WHERE NAME=\'{job_name}\'")
        except:
            self._connect.rollback()    # RollBack if failure to delete

    def list_jobs(self):
        """ List all jobs in the DB. """
        self._cursor.execute('SELECT * FROM JOB_INFO')
        return self._cursor.fetchall()



import copy
from typing import *


class Edge:
    """
    A simple container for edges in a graph

    Attributes:
        start(int) : start point of the edge
        end(int) : end point of the edge
        valid(bool) : attribute of the edge indicating whether it's valid
        color(int) : color of the edge
        next(int) : next fake pointer in edge list
    """

    def __init__(self, start: int, end: int, valid: bool = True, color: int = -1, _next: int = -1):
        """
        Construct an edge
        """
        self.start = start
        self.end = end
        self.valid = valid
        self.color = color
        self.next = _next
        """next edge(which is used in head array)"""

    def __repr__(self):
        return f"edge [start: {self.start}," \
               f" end: {self.end}, " \
               f"valid: {self.valid}, " \
               f"color: {self.color}, " \
               f"next={self.next}]"


class Graph:
    """
    A simple container for graphs. The graph is directed.

    For undirected graph, you must add reverse edges just after
    adding an edge. For example, add_edge(a,b), add_edge(b, a).
    Therefore, `eid` of reverse edge is exactly `eid^1`.
    """

    def __init__(self, nodes: List[int] = None, edges: List[Edge] = None):
        """
        Construct a graph. It's empty if both nodes and edges are None.

        Args:
            nodes(List[int]) : A list of all nodes. In most cases, it should be a continuous list
                which contains sequential integers.
            edges(List[Edge]) : A list of all edges
        """
        self.nodes = [] if nodes is None else copy.copy(nodes)
        self.edges = []
        self.head = {}
        for i in self.nodes:
            self.head[i] = -1
        """get edges that share the same head(start point)"""
        if edges is None:
            return
        cnt = 0
        for edge in edges:
            s = edge.start
            e = edge.end
            m = edge.valid
            self.edges.append(Edge(s, e, m, self.head[s]))
            self.head[s] = cnt
            cnt += 1

    def get_degree(self, x: int):
        """
        Get degree of node x

        Args:
            x(int) : Index of node to be queried.

        Returns:
            int : degree of node x.

        """
        deg = 0
        i = self.head[x]
        while i != -1:
            deg += 1
            i = self.edges[i].next
        return deg

    def get_any_degree(self) -> int:
        """
        Return degree of any node. User should not expect return value of
        this function to be special unless it's a regular graph.

        Returns:
            int : degree of any node.
        """
        return self.get_degree(self.nodes[0])

    def get_max_degree(self) -> int:
        """
        Get maximum degree of the graph

        Returns:
            int : max degree
        """
        deg = -1
        for i in self.nodes:
            deg = max(deg, self.get_degree(i))
        return deg

    def get_max_degree_node(self) -> int:
        """
        Get the index of node with maximal degree.
        Returns:
            int : Node index.
        """
        index = -1
        deg = -1
        for node in self.nodes:
            new_deg = self.get_degree(node)
            if deg < new_deg:
                deg = new_deg
                index = node
        return index

    def list_degree(self) -> None:
        s = 0
        for node in self.nodes:
            d = self.get_degree(node)
            print(d)
            s += d
        print(f"summation of degree: {s}")

    def add_edge(self, start: int, end: int, valid: bool = True, color: int = -1):
        """
        Add an edge from start end with marked value

        Args:
            start(int) : start point
            end(int) : end point
            valid(bool) : if valid
            color(int) : color of the edge
        """
        self.edges.append(Edge(start=start, end=end, valid=valid, color=color, _next=self.head[start]))
        self.head[start] = len(self.edges) - 1


class Bipartite(Graph):
    """
    A bipartite graph

    Attributes:
        left(List) : left side of bipartite graph
        right(List) : right side of bipartite graph
    """

    def __init__(self, left: List[int], right: List[int], edges: List[Edge] = None):
        """
        Construct a bipartite graph. There should be no overlap between left and right

        Args:
            left(List[int]) : left side of bipartite graph
            right(List[int]) : right side of bipartite graph
            edges(List[Edge]) : Specify edge between left and right. The next field of Edge could be invalid here.
        """
        self.left = copy.copy(left)
        self.right = copy.copy(right)
        super().__init__(left + right, edges)

    def list_degree(self) -> None:
        s = 0
        print("left side:")
        for node in self.left:
            d = self.get_degree(node)
            print(d)
            s += d
        print("right side:")
        for node in self.right:
            d = self.get_degree(node)
            print(d)
            s += d
        print(f"summation of degree: {s}")



from .. import InstructionSet
from ..transform_rule import xyx_rule

from QuICT.core.gate import *

USTCSet = InstructionSet(
    GateType.cx,
    [GateType.rx, GateType.ry, GateType.rz, GateType.h, GateType.x]
)
USTCSet.register_one_qubit_rule(xyx_rule)



from QuICT.qcda.synthesis.arithmetic.vbe import VBEMulAddMod
from QuICT.qcda.synthesis.arithmetic.bea import BEAMulMod
from QuICT.qcda.synthesis.arithmetic.hrs import HRSMulMod
from QuICT.qcda.synthesis.arithmetic.tmvh import Multiplication



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2021/6/28 下午4:50
# @Author  : Kaiqi Li
# @File    : constant_statevecto_simulator

from typing import Union
import numpy as np
import cupy as cp

from QuICT.core import Circuit
from QuICT.core.operator import Trigger
from QuICT.core.gate import BasicGate, CompositeGate
from QuICT.core.utils import GateType, MatrixType
from QuICT.ops.utils import LinAlgLoader
from QuICT.ops.gate_kernel import float_multiply, complex_multiply
from QuICT.simulation.utils import GateMatrixs
from QuICT.tools.exception.core import ValueError, TypeError, GateQubitAssignedError
from QuICT.tools.exception.simulation import (
    SampleBeforeRunError, GateTypeNotImplementError, StateVectorUnmatchedError, GateAlgorithmNotImplementError
)


class ConstantStateVectorSimulator:
    """
    The simulator for qubits' vector state.

    Args:
        circuit (Circuit): The quantum circuit.
        precision (str): The precision for the state vector, one of [single, double]. Defaults to "double".
        gpu_device_id (int): The GPU device ID.
        matrix_aggregation (bool): Using quantum gate matrix's aggregation to optimize running speed.
        sync (bool): Sync mode or Async mode.
    """
    __PRECISION = ["single", "double"]

    @property
    def circuit(self):
        return self._circuit

    @circuit.setter
    def circuit(self, circuit):
        self._circuit = circuit

    @property
    def vector(self):
        return self._vector

    @vector.setter
    def vector(self, vec):
        with cp.cuda.Device(self._device_id):
            if type(vec) is np.ndarray:
                self._vector = cp.array(vec)
            else:
                self._vector = vec

    @property
    def device(self):
        return self._device_id

    def __init__(
        self,
        precision: str = "double",
        gpu_device_id: int = 0,
        matrix_aggregation: bool = True,
        sync: bool = True
    ):
        if precision not in self.__PRECISION:
            raise ValueError("StateVectorSimulation.precision", "[single, double]", precision)

        self._precision = np.complex128 if precision == "double" else np.complex64
        self._device_id = gpu_device_id
        self._sync = sync

        # Optimization
        self._using_matrix_aggregation = matrix_aggregation

        # Initial simulator with limit_qubits
        self._algorithm = LinAlgLoader(device="GPU", enable_gate_kernel=True, enable_multigpu_gate_kernel=False)
        # Set gpu id
        cp.cuda.runtime.setDevice(self._device_id)

    def initial_circuit(self, circuit: Circuit):
        """ Initial the qubits, quantum gates and state vector by given quantum circuit. """
        self._circuit = circuit
        self._qubits = int(circuit.width())
        self._pipeline = []

        if self._precision != circuit._precision:
            circuit.convert_precision()

        self._pipeline = circuit.gates

        # Initial GateMatrix if matrix_aggregation is True
        if self._using_matrix_aggregation:
            self.gateM_optimizer = GateMatrixs(self._precision, self._device_id)
            self.gateM_optimizer.build(self._pipeline)

    def _get_gate_matrix(self, gate: BasicGate):
        """ Return the gate's matrix in GPU. """
        if self._using_matrix_aggregation:
            return self.gateM_optimizer.get_target_matrix(gate)
        else:
            return cp.array(gate.matrix, dtype=self._precision)

    def initial_state_vector(self, all_zeros: bool = False):
        """ Initial qubits' vector states. """
        vector_size = 1 << int(self._qubits)
        self._vector = cp.zeros(vector_size, dtype=self._precision)
        if not all_zeros:
            self._vector.put(0, self._precision(1))

    def run(
        self,
        circuit: Circuit,
        state_vector: Union[np.ndarray, cp.ndarray] = None,
        use_previous: bool = False,
        gpu_out: bool = True
    ) -> np.ndarray:
        """ start simulator with given circuit

        Args:
            circuit (Circuit): The quantum circuits.
            state_vector (ndarray): The initial state vector.
            use_previous (bool, optional): Using the previous state vector. Defaults to False.
            record_runtime_measured (bool, optional): Record qubits' measured state during the simulation.

        Returns:
            [array]: The state vector.
        """
        self.initial_circuit(circuit)
        if state_vector is not None:
            assert 2 ** self._qubits == state_vector.size, \
                StateVectorUnmatchedError("The state vector should has the same qubits with the circuit.")
            self.vector = cp.array(state_vector, dtype=self._precision)
        elif not use_previous:
            self.initial_state_vector()

        idx = 0
        while idx < len(self._pipeline):
            gate = self._pipeline[idx]
            idx += 1
            if isinstance(gate, BasicGate):
                self.apply_gate(gate)
            elif isinstance(gate, Trigger):
                self.apply_trigger(gate, idx)
            else:
                raise TypeError("StateVectorSimulation.run.circuit", "[BasicGate, Trigger]". type(gate))

        if not gpu_out:
            return self.vector.get()
        else:
            return self.vector

    def apply_gate(self, gate: BasicGate):
        """ Depending on the given quantum gate, apply the target algorithm to calculate the state vector.

        Args:
            gate (Gate): the quantum gate in the circuit.
        """
        matrix_type = gate.matrix_type
        gate_type = gate.type
        default_parameters = (self._vector, self._qubits, self._sync)
        if (
            gate_type == GateType.id or
            gate_type == GateType.barrier or
            gate.is_identity()
        ):
            return

        # Deal with quantum gate with more than 3 qubits.
        if (
            gate_type in [GateType.unitary, GateType.qft, GateType.iqft, GateType.perm_fx] and
            gate.targets >= 3
        ):
            matrix = self._get_gate_matrix(gate)
            matrix = matrix.reshape(1 << (gate.controls + gate.targets), 1 << (gate.controls + gate.targets))
            self._vector = self._algorithm.matrix_dot_vector(
                self._vector,
                self._qubits,
                matrix,
                gate.cargs + gate.targs,
                self._sync
            )

            return

        # [H, Hy, SX, SY, SW, U2, U3, Rx, Ry] 2-bits [CH, ] 2-bits[Rzx, targets, unitary]
        if matrix_type == MatrixType.normal:
            self.apply_normal_matrix(gate)
        # [Rz, Phase], 2-bits [CRz, Rzz], 3-bits [CCRz]
        elif matrix_type in [MatrixType.diagonal, MatrixType.diag_diag]:
            self.apply_diagonal_matrix(gate)
        # [X] 2-bits [swap, iswap, iswapdg, sqiswap] 3-bits [CSWAP]
        elif matrix_type == MatrixType.swap:
            self.apply_swap_matrix(gate)
        # [Y] 2-bits [CX, CY] 3-bits: [CCX]
        elif matrix_type == MatrixType.reverse:
            self.apply_reverse_matrix(gate)
        # [S, sdg, Z, U1, T, tdg] # 2-bits [CZ, CU1]
        elif matrix_type == MatrixType.control:
            self.apply_control_matrix(gate)
        # [FSim]
        elif matrix_type == MatrixType.ctrl_normal:
            t_indexes = [self._qubits - 1 - targ for targ in gate.targs]
            matrix = self._get_gate_matrix(gate)
            self._algorithm.ctrl_normal_targs(
                t_indexes,
                matrix,
                *default_parameters
            )
        # [Rxx, Ryy]
        elif matrix_type == MatrixType.normal_normal:
            t_indexes = [self._qubits - 1 - targ for targ in gate.targs]
            matrix = self._get_gate_matrix(gate)
            self._algorithm.normal_normal_targs(
                t_indexes,
                matrix,
                *default_parameters
            )
        # [Rzx]
        elif matrix_type == MatrixType.diag_normal:
            t_indexes = [self._qubits - 1 - targ for targ in gate.targs]
            matrix = self._get_gate_matrix(gate)
            self._algorithm.diagonal_normal_targs(
                t_indexes,
                matrix,
                *default_parameters
            )
        # [Measure, Reset]
        elif gate_type in [GateType.measure, GateType.reset]:
            index = self._qubits - 1 - gate.targ
            prob = self.get_measured_prob(index).get()
            self.apply_specialgate(index, gate_type, prob)
        # [Perm]
        elif gate_type == GateType.perm:
            args = gate.cargs + gate.targs
            if len(args) == self._qubits:
                mapping = np.array(gate.pargs, dtype=np.int32)
            else:
                mapping = np.arange(self._qubits, dtype=np.int32)
                for idx, parg in enumerate(gate.pargs):
                    mapping[args[idx]] = args[parg]

            self._algorithm.VectorPermutation(
                self._vector,
                mapping,
                changeInput=True,
                gpu_out=False,
                sync=self._sync
            )
        # unsupported quantum gates
        else:
            raise GateTypeNotImplementError(f"Unsupported Gate Type and Matrix Type: {gate_type} {matrix_type}.")

    def apply_normal_matrix(self, gate: BasicGate):
        # Get gate's parameters
        args_num = gate.controls + gate.targets
        gate_args = gate.cargs + gate.targs
        matrix = self._get_gate_matrix(gate)
        default_parameters = (self._vector, self._qubits, self._sync)

        # Deal with 1-qubit normal gate e.g. H
        if args_num == 1:
            index = self._qubits - 1 - gate_args[0]
            self._algorithm.normal_targ(
                index,
                matrix,
                *default_parameters
            )
        elif args_num == 2:     # Deal with 2-qubits control normal gate e.g. CH
            if gate.controls == gate.targets:
                c_index = self._qubits - 1 - gate.carg
                t_index = self._qubits - 1 - gate.targ
                self._algorithm.normal_ctargs(
                    c_index,
                    t_index,
                    matrix,
                    *default_parameters
                )
            elif gate.targets == 2:     # Deal with 2-qubits unitary gate
                indexes = [self._qubits - 1 - index for index in gate_args]
                self._algorithm.normal_targs(
                    indexes,
                    matrix,
                    *default_parameters
                )
            else:
                raise GateQubitAssignedError("Quantum gate cannot only have control qubits.")

    def apply_diagonal_matrix(self, gate: BasicGate):
        # Get gate's parameters
        args_num = gate.controls + gate.targets
        gate_args = gate.cargs + gate.targs
        matrix = self._get_gate_matrix(gate)
        default_parameters = (self._vector, self._qubits, self._sync)

        # Deal with 1-qubit diagonal gate, e.g. Rz
        if args_num == 1:
            index = self._qubits - 1 - gate.targ
            self._algorithm.diagonal_targ(
                index,
                matrix,
                *default_parameters
            )
        elif args_num == 2:     # Deal with 2-qubit diagonal gate, e.g. CRz
            if gate.controls == gate.targets:
                c_index = self._qubits - 1 - gate.carg
                t_index = self._qubits - 1 - gate.targ
                self._algorithm.diagonal_ctargs(
                    c_index,
                    t_index,
                    matrix,
                    *default_parameters
                )
            elif gate.targets == 2:
                indexes = [self._qubits - 1 - index for index in gate_args]
                self._algorithm.diagonal_targs(
                    indexes,
                    matrix,
                    *default_parameters
                )
            else:
                raise GateQubitAssignedError("Quantum gate cannot only have control qubits.")
        else:   # [CCRz]
            assert gate.type == GateType.ccrz, \
                GateAlgorithmNotImplementError(f"State Vector Simulator cannot deal with {gate.type} currently.")
            c_indexes = [self._qubits - 1 - carg for carg in gate.cargs]
            t_index = self._qubits - 1 - gate.targ
            self._algorithm.diagonal_more(
                c_indexes,
                t_index,
                matrix,
                *default_parameters
            )

    def apply_swap_matrix(self, gate: BasicGate):
        # Get gate's parameters
        args_num = gate.controls + gate.targets
        gate_args = gate.cargs + gate.targs
        default_parameters = (self._vector, self._qubits, self._sync)

        if args_num == 1:       # Deal with X Gate
            index = self._qubits - 1 - gate.targ
            self._algorithm.swap_targ(
                index,
                *default_parameters
            )
        elif args_num == 2:     # Deal with Swap Gate
            if gate.targets != 2:
                raise GateAlgorithmNotImplementError(f"State Vector Simulator cannot deal with {gate.type} currently.")

            t_indexes = [self._qubits - 1 - targ for targ in gate_args]
            self._algorithm.swap_targs(
                t_indexes,
                self._get_gate_matrix(gate),
                *default_parameters
            )
        else:   # CSwap
            assert gate.type == GateType.cswap, \
                GateAlgorithmNotImplementError(f"State Vector Simulator cannot deal with {gate.type} currently.")
            t_indexes = [self._qubits - 1 - targ for targ in gate.targs]
            c_index = self._qubits - 1 - gate.carg
            self._algorithm.swap_tmore(
                t_indexes,
                c_index,
                *default_parameters
            )

    def apply_reverse_matrix(self, gate: BasicGate):
        # Get gate's parameters
        args_num = gate.controls + gate.targets
        gate_args = gate.cargs + gate.targs
        matrix = self._get_gate_matrix(gate)
        default_parameters = (self._vector, self._qubits, self._sync)

        if args_num == 1:   # Deal with 1-qubit reverse gate, e.g. Y
            index = self._qubits - 1 - gate_args[0]
            self._algorithm.reverse_targ(
                index,
                matrix,
                *default_parameters
            )
        elif args_num == 2:   # only consider 1 control qubit + 1 target qubit
            if gate.targets == 2:
                raise GateAlgorithmNotImplementError(f"State Vector Simulator cannot deal with {gate.type} currently.")

            c_index = self._qubits - 1 - gate_args[0]
            t_index = self._qubits - 1 - gate_args[1]
            self._algorithm.reverse_ctargs(
                c_index,
                t_index,
                matrix,
                *default_parameters
            )
        else:   # CCX
            assert gate.type == GateType.ccx, \
                GateAlgorithmNotImplementError(f"State Vector Simulator cannot deal with {gate.type} currently.")
            c_indexes = [self._qubits - 1 - carg for carg in gate.cargs]
            t_index = self._qubits - 1 - gate.targ
            self._algorithm.reverse_more(
                c_indexes,
                t_index,
                *default_parameters
            )

    def apply_control_matrix(self, gate: BasicGate):
        # Get gate's parameters
        args_num = gate.controls + gate.targets
        gate_args = gate.cargs + gate.targs
        default_parameters = (self._vector, self._qubits, self._sync)

        if args_num == 1:       # Deal with 1-qubit control gate, e.g. S
            index = self._qubits - 1 - gate_args[0]
            val = gate.matrix[1, 1]
            self._algorithm.control_targ(
                index,
                val,
                *default_parameters
            )
        elif args_num == 2:     # Deal with 2-qubit control gate, e.g. CZ
            if gate.targets == 2:
                raise GateAlgorithmNotImplementError(f"State Vector Simulator cannot deal with {gate.type} currently.")

            c_index = self._qubits - 1 - gate_args[0]
            t_index = self._qubits - 1 - gate_args[1]
            val = gate.matrix[3, 3]
            self._algorithm.control_ctargs(
                c_index,
                t_index,
                val,
                *default_parameters
            )
        else:
            raise GateAlgorithmNotImplementError(f"State Vector Simulator cannot deal with {gate.type} currently.")

    def apply_trigger(self, op: Trigger, current_idx: int) -> CompositeGate:
        """ Deal with the Operator <Trigger>.

        Args:
            op (Trigger): The operator Trigger
            current_idx (int): the index of Trigger in Circuit
        """
        for targ in op.targs:
            index = self._qubits - 1 - targ
            prob = self.get_measured_prob(index).get()
            _ = self.apply_specialgate(index, GateType.measure, prob)

        mapping_cgate = op.mapping(int(self._circuit[op.targs]))
        if isinstance(mapping_cgate, CompositeGate):
            # optimized composite gate's matrix
            if self._using_matrix_aggregation:
                self.gateM_optimizer.build(mapping_cgate.gates)

            # Check for checkpoint
            cp = mapping_cgate.checkpoint
            position = current_idx if cp is None else self._circuit.find_position(cp)
            self._pipeline = self._pipeline[:position] + mapping_cgate.gates + \
                self._pipeline[position:]

    def apply_multiply(self, value: Union[float, np.complex]):
        """ Deal with Operator <Multiply>

        Args:
            value (Union[float, complex]): The multiply value apply to state vector.
        """
        default_parameters = (self._vector, self._qubits, self._sync)
        if isinstance(value, float):
            float_multiply(value, *default_parameters)
        else:
            if self._precision == np.complex64:
                value = np.complex64(value)

            complex_multiply(value, *default_parameters)

    def apply_zeros(self):
        """ Set state vector to be zero. """
        self._vector = cp.zeros_like(self.vector)

    def get_measured_prob(self, index: int, all_measured: bool = False) -> cp.ndarray:
        """ Return the probability of measured qubit with given index to be 0

        Args:
            index (int): The given qubit index
            all_measured (bool): Calculate measured probability with all state vector,
                only using with Multi-Node Simulation.
        """
        return self._algorithm.measured_prob_calculate(
            index,
            self._vector,
            self._qubits,
            all_measured=all_measured,
            sync=self._sync
        )

    def apply_specialgate(self, index: int, type: GateType, prob: float = None):
        """ Apply Measure/Reset gate in to simulator

        Args:
            index (int): The given qubit index
            type (GateType): the gate type of special gate
            prob (float): The given probability of measured the target qubit into 1

        Returns:
            [float]: The target qubit's measured value or reset value, <0 or <1
        """
        if type == GateType.measure:
            result = int(self._algorithm.apply_measuregate(
                index,
                self._vector,
                self._qubits,
                prob,
                self._sync
            ))
            self._circuit.qubits[self._qubits - 1 - index].measured = result
        elif type == GateType.reset:
            result = self._algorithm.apply_resetgate(
                index,
                self._vector,
                self._qubits,
                prob,
                self._sync
            )
        else:
            raise ValueError("StateVectorSimulator.apply_specialgate", "[measure, reset]", type)

        return result

    def sample(self, shots: int = 1) -> list:
        """ Sample the measured result from current state vector, please initial Circuit first

        Args:
            shots (int): The sample times of current state vector.

        Returns:
            List[int]: The measured result list with length equal to 2 ** self.qubits
        """
        assert (self._circuit is not None), \
            SampleBeforeRunError("StateVectorSimulation sample without run any circuit.")
        original_sv = self._vector.copy()
        state_list = [0] * (1 << self._qubits)
        lastcall_per_qubit = self._circuit.get_lastcall_for_each_qubits()
        measured_idx = [
            i for i in range(self._qubits)
            if lastcall_per_qubit not in [GateType.reset, GateType.measure]
        ]

        for _ in range(shots):
            for m_id in measured_idx:
                index = self._qubits - 1 - m_id
                prob = self.get_measured_prob(index).get()
                _ = self.apply_specialgate(index, GateType.measure, prob)

            state_list[int(self._circuit.qubits)] += 1
            self._vector = original_sv.copy()

        return state_list



#!/home/likaiqi/.conda/envs/env_kq/bin/python

from argparse import ArgumentParser, Namespace, RawTextHelpFormatter
from copy import deepcopy


QUICT_DESCRIBE = """
   ___            ___    ____   _____
  / _ \   _   _  |_ _|  / ___| |_   _|
 | | | | | | | |  | |  | |       | |
 | |_| | | |_| |  | |  | |___    | |
  \__\_\  \___/  |___|  \____|   |_|


Welcome to QuICT Command Line Interface.

Using 'quict --help' to get the documents about QuICT CLI.
Using 'quict --version' to get the version of QuICT.
"""


def cli_construct():
    parser = ArgumentParser(
        prog="quict",
        description=QUICT_DESCRIBE,
        formatter_class=RawTextHelpFormatter
    )
    parser.add_argument('--version', action='version', version='QuICT CLI 1.0', help="Show the current version.")
    subparsers = parser.add_subparsers()

    # Build env management
    from QuICT.tools.cli.blueprint.remote import login, logout, register, unsubscribe

    # Login
    login_sp = subparsers.add_parser(
        name="login",
        description="Login to QuICT Cloud System.",
        help="Login to QuICT Cloud System.",
    )
    login_sp.add_argument(
        "username", type=str,
        help="The username used to login."
    )
    login_sp.add_argument(
        "password", type=str,
        help="The password of user."
    )
    login_sp.set_defaults(func=login)

    # Logout
    logout_sp = subparsers.add_parser(
        name="logout",
        description="Logout from the QuICT.",
        help="Logout from the QuICT."
    )
    logout_sp.set_defaults(func=logout)

    # Register
    register_sp = subparsers.add_parser(
        name="register",
        description="User Register for QuICT Cloud System.",
        help="User Register",
    )
    register_sp.add_argument(
        "username", type=str,
        help="The username used."
    )
    register_sp.add_argument(
        "password", type=str,
        help="The password for user."
    )
    register_sp.add_argument(
        "email", type=str,
        help="The email address for registed user."
    )
    register_sp.set_defaults(func=register)

    # unsubscribe
    unsubscribe_sp = subparsers.add_parser(
        name="unsubscribe",
        description="Unsubscribe user to QuICT Cloud System.",
        help="Unsubscribe to QuICT Cloud System.",
    )
    unsubscribe_sp.add_argument(
        "username", type=str,
        help="The username used to unsubscribe."
    )
    unsubscribe_sp.add_argument(
        "password", type=str,
        help="The password of user."
    )
    unsubscribe_sp.set_defaults(func=unsubscribe)

    # Circuit
    circuit_sp = subparsers.add_parser(
        name="circuit",
        description="Manage the local circuit library.",
        help="get quantum circuit qasm."
    )
    circuit_cli_construct(circuit_sp)

    # Local Mode's Job
    local_sp = subparsers.add_parser(
        name="local",
        description="Local Modes QuICT Job Management",
        help="QuICT job's management in Local Mode."
    )
    job_cli_construct(local_sp, mode="local")

    # Local Mode's Job
    remote_sp = subparsers.add_parser(
        name="remote",
        description="Remote Modes QuICT Job Management.",
        help="QuICT job's management in Remote Mode."
    )
    job_cli_construct(remote_sp, mode="remote")

    # Benchmark
    benchmark_sp = subparsers.add_parser(
        name="benchmark",
        description="QuICT Benchmark",
        help="QuICT benchmark"
    )
    benchmark_cli_construct(benchmark_sp)

    return parser.parse_args()


def circuit_cli_construct(circuit_sp: ArgumentParser):
    """ Build circuit module in CLI

    Args:
        circuit_sp (ArgumentParser): Circuit Parser
    """
    from QuICT.tools.cli.blueprint.circuit import (
        get_random_circuit, get_algorithm_circuit, store_quantum_circuit,
        delete_quantum_circuit, list_quantum_circuit
    )

    subparser = circuit_sp.add_subparsers()
    # quict circuit get_random
    get_random = subparser.add_parser(
        name="get_random",
        description="Get random circuit.",
        help="get random circuit"
    )
    get_random.add_argument(
        "-i", "--instruction_set",
        choices=["ctrl_unitary", "diag", "single_bit", "ctrl_diag", "google", "ibmq", "ionq", "ustc", "nam", "origin"],
        nargs="?", default="nam",
        help="Choice the instrcution set for random circuit."
    )
    get_random.add_argument(
        "-q", "--qubits",
        nargs="+", type=int, default=[5],
        help="The number of qubits."
    )
    get_random.add_argument(
        "-ms", "--max_size",
        type=int, default=None,
        help="The maximum number of quantum gates."
    )
    get_random.add_argument(
        "-md", "--max_depth",
        type=int, default=None,
        help="The maximum number of circuit depth."
    )
    get_random.add_argument(
        "output_path",
        nargs="?", default=".",
        help="The output path, default to be current path."
    )
    get_random.set_defaults(func=get_random_circuit)

    # quict circuit get_algorithm
    get_algorithm = subparser.add_parser(
        name="get_algorithm",
        description="Get quantum algorithm's circuit.",
        help="get quantum algorithm's circuit, for some algorithm, may have qubits limitation."
    )
    get_algorithm.add_argument(
        "alg", nargs="?",
        choices=["adder", "clifford", "cnf", "grover", "maxcut", "qft", "qnn", "quantum_walk", "vqe"], default="qft",
        help="The quantum algorithm."
    )
    get_algorithm.add_argument(
        "-q", "--qubits",
        nargs="+", type=int, default=5,
        help="The number of qubits' number."
    )
    get_algorithm.add_argument(
        "output_path",
        nargs="?", default=".",
        help="The output path, default to be current path."
    )
    get_algorithm.set_defaults(func=get_algorithm_circuit)

    # quict circuit add
    add = subparser.add_parser(
        name="add",
        description="Store quantum circuit qasm.",
        help="store quantum circuit qasm"
    )
    add.add_argument(
        "name", type=str,
        help="The name of quantum circuit."
    )
    add.add_argument(
        "file", nargs="?",
        type=str, default=".",
        help="The path of qasm file."
    )
    add.set_defaults(func=store_quantum_circuit)

    # quict circuit delete
    delete = subparser.add_parser(
        name="delete",
        description="delete quantum circuit.",
        help="delete the quantum circuit"
    )
    delete.add_argument(
        "name", type=str,
        help="The name of quantum circuit."
    )
    delete.set_defaults(func=delete_quantum_circuit)

    # quict circuit list
    list_cir = subparser.add_parser(
        name="list",
        description="list all customed quantum circuit.",
        help="list all customed quantum circuit."
    )
    list_cir.set_defaults(func=list_quantum_circuit)


def job_cli_construct(mode_sp: ArgumentParser, mode: str):
    """ Build Job module [include remote and local mode] in CLI

    Args:
        mode_sp (ArgumentParser): Job Mode Parser
        mode (str): mode description, one of [local, remote]
    """
    from QuICT.tools.cli.blueprint.job import get_template

    if mode == "local":
        from QuICT.tools.cli.blueprint.job import (
            start_job, stop_job, restart_job, delete_job, status_job, list_jobs
        )
    elif mode == "remote":
        from QuICT.tools.cli.blueprint.remote import start_job, delete_job, status_job, list_jobs

    mode_subparser = mode_sp.add_subparsers()
    job_sp = mode_subparser.add_parser(
        name="job",
        description="QuICT Jobs Management.",
        help="QuICT job's related tools."
    )

    # quict job get_template
    subparser = job_sp.add_subparsers()
    get_templates = subparser.add_parser(
        name="get_template",
        description="Get job template.",
        help="Get job's template."
    )
    get_templates.add_argument(
        "output_path", nargs="?", default=".",
        help="The output path, default to be current path."
    )
    get_templates.set_defaults(func=get_template)

    # quict job start
    start = subparser.add_parser(
        name="start",
        description="Start the job.",
        help="start the job."
    )
    start.add_argument(
        "file", type=str,
        help="The path of jobs file, could be a directory or some file path.",
    )
    start.set_defaults(func=start_job)

    # quict job status
    status = subparser.add_parser(
        name="status",
        description="check the job's status.",
        help="check the job."
    )
    status.add_argument(
        "name", type=str,
        help="The name of target job."
    )
    status.set_defaults(func=status_job)

    if mode == "local":
        # quict job stop
        stop = subparser.add_parser(
            name="stop",
            description="stop a job.",
            help="stop a job."
        )
        stop.add_argument(
            "name", type=str,
            help="The name of target job."
        )
        stop.set_defaults(func=stop_job)

        # quict job restart
        restart = subparser.add_parser(
            name="restart",
            description="restart a stopped job.",
            help="restart a stopped job."
        )
        restart.add_argument(
            "name", type=str,
            help="The name of target job."
        )
        restart.set_defaults(func=restart_job)

    # quict job delete
    delete = subparser.add_parser(
        name="delete",
        description="delete the job.",
        help="delete the job."
    )
    delete.add_argument(
        "name", type=str,
        help="The name of target job."
    )
    delete.set_defaults(func=delete_job)

    # quict job list
    list_job = subparser.add_parser(
        name="list",
        description="list all jobs.",
        help="list all jobs."
    )
    list_job.set_defaults(func=list_jobs)


def benchmark_cli_construct(benchmark_sp: ArgumentParser):
    """ Build Benchmark Module in CLI

    Args:
        benchmark_sp (ArgumentParser): Benchmark Parser
    """
    from QuICT.tools.cli.blueprint.benchmark import benchmark

    benchmark_sp.add_argument(
        "--gpu", action="store_true",
        help="The name of target job."
    )
    benchmark_sp.set_defaults(func=benchmark)


def _decompose_namespace(args: Namespace):
    mapping_args = vars(deepcopy(args))
    del mapping_args["func"]

    return mapping_args


def main():
    args = cli_construct()
    map_args = _decompose_namespace(args)
    args.func(**map_args)


if __name__ == "__main__":
    main()



ERROR_CODE = {
    # Error code table for the QuICT system.
    1000: "QuICT Exception",
    # 1001-1999: QuICT Core Module Exception
    1001: "QuICT Core Module Type Exception",
    1002: "QuICT Core Module Value Exception",
    1003: "QuICT Core Module Index Exceed Exception",
    1004: "QuICT Circuit Append Operator Exception",
    1005: "QuICT Circuit Replace Operator Exception",
    1006: "QuICT Circuit Random or Supremacy Append Exception",
    1007: "QuICT DAG Circuit Exception",
    1008: "QuICT Circuit Draw Exception",
    1009: "QuICT QASM Exception",
    1010: "QuICT Qubit Measure Exception",
    1011: "QuICT CheckPointChildren no Find Exception",
    1012: "QuICT Noise Kraus Matrix Exception",
    1013: "QuICT Noise Apply Exception",
    1014: "QuICT Noise Pauli Operators Unmatched Exception",
    1015: "QuICT Noise Dampling Prob Exceed Exception",
    1016: "QuICT Gate Qubit Assigned Exception",
    1017: "QuICT Gate Parameter Exception",
    1018: "QuICT Gate Append Exception",
    1019: "QuICT Gate Matrix Exception",
    1020: "QuICT CompositeGate Append Exception",

    # 2000-2999: QuICT Simulation Module Exception
    2001: "QuICT Simulation Sample before Run Exception",
    2002: "QuICT Simulation Gate not Implement Exception",
    2003: "QuICT Simulation Matrix Exception",
    2004: "QuICT Simulation State Vector unmatched Exception",
    2005: "QuICT Simulation Gate Algorithm not Implement Exception",
    2006: "QuICT Simulation Unitary Matrix Unmatched Exception",
    2007: "QuICT Simulation Options Unmatched Exception",

    # 3000-3999: QuICT Algorithm Module Exception

    # 4000-4999: QuICT QCDA Module Exception

    # 9000-9999: QuICT Utility Module Exception

}


class QuICTException(Exception):
    """The base exception class for QuICT.

    Args:
        error_code (int): the predefined QuICT error code.
        msg (str): the description of the error. If is None, which will
        show the base error information.
    """

    def __init__(self, error_code: int = 1000, msg: str = None):
        self.ecode = error_code
        self.strerror = msg if msg else ERROR_CODE[self.ecode]
        if not isinstance(self.strerror, str):
            self.strerror = str(msg)

    def __str__(self):
        return self.strerror

    def __repr__(self):
        return f"{self.__class__.__name__}({str(self)})"



import numpy as np
import cupy as cp

from QuICT.core.operator import DataSwitch, DataSwitchType


class DataSwitcher:
    """ A class of data switch functions using by multi-GPU simulator.

    Args:
        proxy (Proxy): The NCCL Communicator.
        qubits (int): The number of qubits.
    """
    @property
    def id(self):
        return self._id

    def __init__(self, proxy):
        self._proxy = proxy
        self._id = proxy.dev_id

    def __call__(self, op: DataSwitch, vector):
        destination = op.destination
        if op.type == DataSwitchType.all:
            self.all_switch(vector, destination)
        elif op.type == DataSwitchType.half:
            self.half_switch(vector, destination)
        elif op.type == DataSwitchType.ctarg:
            self.ctargs_switch(vector, destination, op.switch_condition)
        else:
            raise TypeError("unsupportted data switch type.")

    def _switch(self, vector, destination: int):
        """ Based data switch function, swithc the data between self and destination.

        Args:
            vector (cp.array): The data which will be sent to destination.
            destination (int): The communicator which switch the data with us.
            _0_1 (bool): Is the initiator or the follower about the switch behavior.

        Returns:
            np.array: the switched data from the destination.
        """
        is_initiator = self._id < destination
        recv_buf = cp.zeros(vector.size, dtype=vector.dtype)

        if is_initiator:
            self._proxy.send(vector, destination)
            self._proxy.recv(recv_buf, destination)
        else:
            self._proxy.recv(recv_buf, destination)
            self._proxy.send(vector, destination)

        return recv_buf

    def all_switch(self, vector, destination: int):
        """ Switch the whole data with the destination.

        Args:
            vector (cp.array): the switched data.
            destination (int): The communicator which switch the data with us.
        """
        vector[:] = self._switch(vector, destination)

    def half_switch(self, vector, destination: int):
        """ Switch the half data with the destination.

        Args:
            vector (cp.array): the switched data.
            destination (int): The communicator which switch the data with us.
        """
        _0_1 = self._id < destination
        sending_size = vector.size // 2

        if not _0_1:
            sending_data = vector[:sending_size]
            recv_buf = self._switch(sending_data, destination)
            vector[:sending_size] = recv_buf
        else:
            sending_data = vector[sending_size:]
            recv_buf = self._switch(sending_data, destination)
            vector[sending_size:] = recv_buf

    def quarter_switch(self, vector, destination: list):
        assert(len(destination) == 4)
        sending_size = vector.size // 4

        for idx, dest in enumerate(destination):
            if dest != self._id:
                sending_data = vector[idx * sending_size: (idx + 1) * sending_size]
                vector[idx * sending_size: (idx + 1) * sending_size] = self._switch(sending_data, dest)

    def ctargs_switch(self, vector, destination: int, condition: dict):
        """ Switch the data by the given condition.
            e.g. if condition = {3: 1}, switch the data which the third bit-indexes is 1 with the data
            from the destination.

        Args:
            vector (cp.array): the switched data.
            destination (int): The communicator which switch the data with us.
            condition (dict[index(int): 0/1]): Describe the indexes of the data will be switched.
        """
        target_idx = np.arange(vector.size, dtype=np.int64)
        for idx, _0_1 in condition.items():
            if isinstance(target_idx, tuple):
                target_idx = target_idx[0]

            if _0_1:
                target_idx = target_idx[np.where(target_idx & (1 << idx))]
            else:
                target_idx = target_idx[np.where((target_idx & (1 << idx)) == 0)]

        sending_data = vector[target_idx]
        recv_buf = self._switch(sending_data, destination)
        vector[target_idx] = recv_buf

    def add_prob(self, prob_result):
        recv_buf = cp.zeros_like(prob_result)

        self._proxy.allreduce(
            sendbuf=prob_result,
            recvbuf=recv_buf,
            op=0
        )

        return recv_buf



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2020/12/27 10:13 下午
# @Author  : Han Yu
# @File    : __init__.py

from .uniformly_rotation import UniformlyRotation
from .uniformly_unitary import UniformlyUnitary
